from __future__ import annotations

import logging
from dataclasses import dataclass
from inspect import signature
from pathlib import Path
from typing import Callable, Optional

from testscribe.file_info import get_function
from testscribe.log import log
from testscribe.module import Module

logger = logging.getLogger(__name__)


def create_output_dir_for_module(output_root_dir: Path, module: Module) -> Path:
    output_dir = output_root_dir.joinpath(*module.get_package_name_list())
    # Make sure the directory exists. If it doesn't exist,
    # create it including the parent directories.
    output_dir.mkdir(parents=True, exist_ok=True)
    return output_dir


def get_scribe_file_path(output_root_dir: Path, module: Module) -> Path:
    """
    This function will ensure the parent directory exists.

    :param output_root_dir:
    :param module:
    :return:
    """
    output_dir = create_output_dir_for_module(
        output_root_dir=output_root_dir, module=module
    )
    module_name = module.get_module_name_only()
    return output_dir.joinpath(f"{module_name}.tscribe")


def get_target_function_name(input_params: InputParams) -> str:
    return input_params.func.__name__


def get_target_class_name(input_params: InputParams) -> str:
    clazz = input_params.clazz
    if clazz:
        return clazz.__qualname__
    else:
        return ""


@dataclass(frozen=True)
class InputParams:
    func: Callable
    output_root_dir: Path
    module: Module
    # if the target is a method, this is the class type of the associated class
    clazz: Optional[type]
    ask_for_test_name: bool
    ask_for_description: bool


def create_input_params(
    module: Module,
    function_name: str,
    output_root_dir: Path,
    ask_for_test_name: bool,
    ask_for_description: bool,
) -> InputParams:
    func, clazz = get_function(module.get_module_str(), function_name)
    sig = signature(func)
    log(f"The target function: {function_name} has the signature: {sig}")
    if clazz:
        log(f"The target class is: {clazz}")
    return InputParams(
        func=func,
        output_root_dir=output_root_dir,
        module=module,
        clazz=clazz,
        ask_for_test_name=ask_for_test_name,
        ask_for_description=ask_for_description,
    )



from __future__ import annotations

import logging
from typing import List, Any

from testscribe import global_var
from testscribe.custom_type import Spec
from testscribe.log import log
from testscribe.mock_call import MockCall
from testscribe.mock_proxy_support import (
    create_mock_name,
    check_unsupported_attributes,
    create_mock_call,
    get_mock_attribute_value,
)
from testscribe.reflection_util import is_method
from testscribe.util import consistent_hash_str
from testscribe.value_util import get_value

logger = logging.getLogger(__name__)


class MockProxy:
    def __init__(self, spec: Spec, name: str = ""):
        # The internal attributes have _test_scribe_ postfix to reduce the chance
        # of collision with attributes of the targets being mocked.
        self.spec_test_scribe_ = spec  # the class or function it is mocking
        self.calls_test_scribe_: List[MockCall] = []
        mock_name = create_mock_name(name=name, spec=spec)
        self.name_test_scribe_ = mock_name
        self.attributes_test_scribe_ = {}
        self.attributes_transformed_test_scribe_ = {}

        global_var.g_name_mock_dict[mock_name] = self
        log(f"Created a mock: {get_proxy_str(self)}")

    def __getattr__(self, name):
        """
        Use __getattribute__ will result in a recursive call into
        __getattribute__ when method_calls attribute is used.
        see
        https://stackoverflow.com/questions/4295678/understanding-the-difference-between-getattr-and-getattribute

        :param name:
        :return:
        """
        logger.debug(f"__getattr__ called with {name}")
        # print_stack()

        check_unsupported_attributes(name)

        if name in self.attributes_test_scribe_:
            # Mocked property value can't change
            v = self.attributes_test_scribe_[name]
            logger.debug(f"Return existing attribute value({v})")
            return v

        if is_method(self.spec_test_scribe_, name):
            return record_mock_call(proxy=self, method_name=name)
        else:
            return record_attribute(proxy=self, attribute_name=name)

    def __call__(self, *args, **kwargs):
        # assume this is a constructor call or a function call
        # todo: support __call__
        mock_call = record_mock_call(proxy=self, method_name="")
        return mock_call.call_internal(*args, **kwargs)

    def __str__(self):
        # __getattr__ is not called with __str__
        call = self.__getattr__("__str__")
        return call.call_internal()

    def __enter__(self):
        # __getattr__ is not called with __enter__ nor __exit__
        # todo: fix the call stack display
        # log(f"__enter__ is called on {self}")
        call = self.__getattr__("__enter__")
        return call.call_internal()

    def __exit__(self, exc_type, exc_value, traceback):
        # log(f"__exit__ is called on {self}.")
        call = self.__getattr__("__exit__")
        return call.call_internal(exc_type, exc_value, traceback)

    def __hash__(self):
        if is_method(self.spec_test_scribe_, "__hash__"):
            # This simple approach doesn't allow the flexibility of mocking __hash__
            # differently. Note that the mock object is transformed into MockName
            # and the __hash__ method may be called when generating the code
            # Let's favor simplity for now.

            # return the hash of self is problematic since the mock object is stored
            # and transformed as MockName. The hash function needs to return
            # consistent results.
            return consistent_hash_str(self.name_test_scribe_)
        else:
            raise TypeError(
                f"The mock target {self.spec_test_scribe_} is not hashable."
            )

    @property
    def __class__(self):
        """
        Override this method to allow the mock object to pass the isinstance check
        if the spec is a class/type.

        Modeled after NonCallableMock.__class__
        @property is important to allow overriding this property.
        :return:
        """
        spec = self.spec_test_scribe_
        if isinstance(spec, type):
            return spec
        else:
            return type(spec)


def get_proxy_str(m: MockProxy):
    return f"Mock: name ({m.name_test_scribe_}) spec ({m.spec_test_scribe_})"


def record_mock_call(proxy: MockProxy, method_name: str) -> MockCall:
    calls = proxy.calls_test_scribe_
    mock_call = create_mock_call(
        method_name=method_name,
        mock_name=proxy.name_test_scribe_,
        spec=proxy.spec_test_scribe_,
        mock_calls=calls,
    )
    calls.append(mock_call)
    return mock_call


def record_attribute(proxy: MockProxy, attribute_name: str) -> Any:
    input_value = get_mock_attribute_value(
        attribute_name=attribute_name,
        mock_name=proxy.name_test_scribe_,
        spec=proxy.spec_test_scribe_,
    )
    # avoid circular reference
    from testscribe.transformer import transform_value

    # keep a copy of the transformed original value
    # in case the value is changed after execution.
    proxy.attributes_transformed_test_scribe_[attribute_name] = transform_value(
        input_value
    )

    value = get_value(input_value)
    proxy.attributes_test_scribe_[attribute_name] = value
    log(f"Mock attribute value: {repr(value)}")

    return value


def is_mock_proxy(obj) -> bool:
    # can't use isinstance on Mock objects because they override __class__
    return issubclass(type(obj), MockProxy)



import logging
from inspect import isfunction, isclass
from pathlib import Path
from typing import List, Optional, Tuple

from testscribe.config import Config
from testscribe.constant import INVALID_TEST_INDEX
from testscribe.error import Error
from testscribe.execution import save_file
from testscribe.execution_util import (
    init,
    get_all_scribe_files,
    compute_output_root_path,
)
from testscribe.file_info import get_module
from testscribe.input_params import (
    get_scribe_file_path,
)
from testscribe.load_scribe_file import load_scribe_file, load_or_create_model
from testscribe.log import log
from testscribe.model_type import TestModel, AllTests, add_test
from testscribe.module import Module
from testscribe.test_name import create_test_name
from testscribe.util import load_object

logger = logging.getLogger(__name__)


def move_tests(
    source_file: Path, class_or_function_name: str, output_root_path: Optional[Path]
):
    """

    :param source_file: the source file that the symbol currently is defined
    :param class_or_function_name: the class or function to move
    :param output_root_path: the root directory of the scribe files to search
    if it is None, use the root path in the config file.
    :return:
    """
    config = init()
    log(f"Move tests that target {class_or_function_name} of {source_file}.")
    output_root_path = compute_output_root_path(
        config=config, output_root_path=output_root_path
    )
    module = get_module(target_file=source_file)
    module_str = module.get_module_str()
    log(f"Target module: {module_str}")
    tests_to_move = get_tests_to_move(
        class_or_function_name=class_or_function_name,
        output_root_path=output_root_path,
        module_str=module_str,
    )
    do_move(
        module=module,
        output_root_path=output_root_path,
        tests_to_move=tests_to_move,
    )


def do_move(module: Module, output_root_path: Path, tests_to_move: List[TestModel]):
    if tests_to_move:
        add_tests(
            output_root_path=output_root_path,
            module=module,
            tests_to_move=tests_to_move,
        )
        return True
    else:
        log("Can't find an existing test targetting this function or class.")
        return False


def get_output_root_path(config: Config, cmd_line_root_path: Optional[Path]) -> Path:
    if cmd_line_root_path:
        log(f"Output root directory cmd line option: {cmd_line_root_path}")
        return cmd_line_root_path
    else:
        return config.output_root_path


def get_tests_to_move(
    class_or_function_name: str, output_root_path: Path, module_str: str
) -> List[TestModel]:
    target_is_class = is_class_name(
        module_str=module_str, class_or_function_name=class_or_function_name
    )
    tests_to_move: List[TestModel] = []
    all_scribe_files = list(get_all_scribe_files(output_root_path))
    log(f"Searching {len(all_scribe_files)} of existing scribe files.")
    logger.debug(f"All scribe files: {all_scribe_files}")

    for scribe_file in all_scribe_files:
        tests = get_tests_to_move_from_one_file(
            class_or_function_name, target_is_class, scribe_file, module_str
        )
        tests_to_move.extend(tests)
    logger.info(f"Found total of {len(tests_to_move)} tests to move.")
    return tests_to_move


def is_class_name(module_str: str, class_or_function_name: str):
    obj = load_object(symbol_name=class_or_function_name, module_str=module_str)
    if isfunction(obj):
        log(f"Target function: {class_or_function_name}")
        return False
    if isclass(obj):
        log(f"Target class: {class_or_function_name}")
        return True
    raise Error(
        f"{class_or_function_name} is not a valid function or class in the module {module_str}"
    )


def get_tests_to_move_from_one_file(
    class_or_function_name: str,
    target_is_class: bool,
    scribe_file: Path,
    target_module_str: str,
) -> List[TestModel]:
    """
    Retrive the existing tests from a given scribe file that need to move.
    If found, it will remove these tests from the existing scribe file.

    If there is an exception when saving the new scribe file, these tests will be
    lost. It's highly suggsted to use a version control system to safeguard against
    such a possibility.

    :param class_or_function_name: the symbol to move
    :param target_is_class: is the target symbol a class name
    :param scribe_file: the scribe file to check
    :param target_module_str: the module to which the symbol now belongs
    :return:
    """
    logger.debug(f"Searching {scribe_file}")
    all_tests = load_scribe_file(scribe_file)
    if should_skip_module(
        module_str=all_tests.module,
        class_or_function_name=class_or_function_name,
        scribe_file=scribe_file,
        target_is_class=target_is_class,
        target_module_str=target_module_str,
    ):
        return []

    tests = all_tests.tests
    remaining_tests, tests_to_move = search_tests_to_remove(
        class_or_function_name=class_or_function_name,
        target_is_class=target_is_class,
        tests=tests,
    )
    if not tests_to_move:
        return []
    log(f"Found {len(tests_to_move)} tests to be moved in {scribe_file}.")
    all_tests.tests = remaining_tests
    # In the rare case when the tests to move fail to save, the information about
    # the existing tests are lost.
    # It's thus highly suggested that the generated scribe files are kept under source
    # control if users want to maintain them.
    save_file(scribe_file, all_tests)
    return tests_to_move


def should_skip_module(
    module_str: str,
    class_or_function_name: str,
    scribe_file: Path,
    target_is_class: bool,
    target_module_str: str,
) -> bool:
    if module_str == target_module_str:
        return True
    if module_contain_same_symbol(
        module_str=module_str,
        name=class_or_function_name,
        is_class=target_is_class,
        target_module_str=target_module_str,
    ):
        log(f"{scribe_file} contains a valid target that has the same name. Skip")
        return True
    return False


def module_contain_same_symbol(
    module_str: str, name: str, is_class: bool, target_module_str: str
) -> bool:
    """
    Return true if the module contains the same symbol name.
    :param module_str: the module to check
    :param name:
    :param is_class: is the name a class name
    :param target_module_str: the module to which the module belong
    :return:
    """
    obj = load_object(symbol_name=name, module_str=module_str)
    if obj is None or obj.__module__ == target_module_str:
        # If the same symbol is imported, it will exist as an attribute of this module.
        # However, the object's __module__ still points to the module where the
        # symbol is defined.
        return False
    if is_class:
        return isclass(obj)
    else:
        return isfunction(obj)


def search_tests_to_remove(
    class_or_function_name: str, target_is_class: bool, tests: List[TestModel]
) -> Tuple[List[TestModel], List[TestModel]]:
    tests_to_move: List[TestModel] = []
    remaining_tests: List[TestModel] = []
    for t in tests:
        if does_test_match_target_name(
            test=t,
            class_or_function_name=class_or_function_name,
            target_is_class=target_is_class,
        ):
            tests_to_move.append(t)
        else:
            remaining_tests.append(t)
    return remaining_tests, tests_to_move


def does_test_match_target_name(
    test: TestModel, class_or_function_name: str, target_is_class: bool
):
    source = get_name_to_compare_with(target_is_class=target_is_class, test=test)
    return source == class_or_function_name


def get_name_to_compare_with(target_is_class: bool, test: TestModel) -> str:
    return test.target_class_name if target_is_class else test.target_func_name


def add_tests(output_root_path: Path, module: Module, tests_to_move: List[TestModel]):
    target_scribe_file_path = get_scribe_file_path(
        output_root_dir=output_root_path, module=module
    )
    log(f"Target scribe file: {target_scribe_file_path}")
    all_tests = load_or_create_model(
        file_path=target_scribe_file_path, full_module_name=module.get_module_str()
    )
    for t in tests_to_move:
        all_tests = add_one_test(all_tests, t)
    save_file(target_scribe_file_path, all_tests)


def add_one_test(all_tests: AllTests, test: TestModel) -> AllTests:
    log(f"Moving test: {test.name}")
    test.name = create_test_name(
        all_tests=all_tests,
        index_of_test_to_update=INVALID_TEST_INDEX,
        short_name=test.short_name,
        target_func_name=test.target_func_name,
    )
    return add_test(all_tests=all_tests, test=test)



DESCRIPTION = "description"
ATTRIBUTES = "attributes"
CALLS = "calls"
SPEC = "spec"
MOCKS = "mocks"
PATCHES = "patches"
RESULT = "result"
EXCEPTION = "exception"
TYPE = "type"
MESSAGE = "message"
RETURN_VALUE = "return"
PARAMETERS = "parameters"
VALUE = "value"
INIT_PARAMETERS = "init_parameters"
TARGET = "target"
CLASS_NAME = "class_name"
NAME = "name"
SHORT_NAME = "short_name"
PATCH_REPLACEMENT = "replacement"
TESTS = "tests"
MODULE = "module"
FORMAT_VERSION = "format_version"



from typing import List, Iterable

from testscribe.config import g_indent
from testscribe.model_type import MockCallModel, ExpressionModel
from testscribe.type_util import is_class_type
from testscribe.util import BUILTIN_MODULE_NAME
from testscribe.value_input_util import eval_special_value


def pretty_format_repr_string(s: str) -> str:
    """
    Use triple double quotes for repr of a multi line string.
    except a string that
    only contains a single '\n' at the end
    or
    contains '\' or triple double quotes

    It's more readable than strings with multiple escape characters.

    :param s:
    :return:
    """
    if "\n" in s[:-1] and not ("\\" in s or '"""' in s):
        if s.endswith('"'):
            # escape the last double quote to avoid a syntax error.
            s = s[:-1] + '\\"'
        format_string = '''\
"""\\
{multi_line_string}"""'''
        return format_string.format(multi_line_string=s)
    return repr(s)


def join_lines(
    lines: List[str], prepend_new_line: bool = True, indentation_level: int = 0
) -> str:
    """
    Join lines with the new line character.
    If there is no line, returns an empty string.

    :param lines:
    :param prepend_new_line: has no effect if there is no line
    :param indentation_level:
    :return:
    """
    assert indentation_level >= 0
    normalized = remove_empty(lines)
    if normalized:
        if indentation_level:
            normalized = add_indentation_to_str_list(normalized, indentation_level)
        statements_str = "\n".join(normalized)
        if prepend_new_line:
            return "\n" + statements_str
        else:
            return statements_str
    else:
        return ""


def is_same_value(values: Iterable) -> bool:
    first = values[0]
    for v in values:
        if v != first:
            return False
    return True


def spec_contain_param_name_info(spec_str: str):
    return not (
        spec_str.startswith("typing.Callable")
        or spec_str.startswith("collections.abc.Callable")
        or spec_str.startswith("collections.Callable")
    )


def get_module_name(v: any) -> str:
    t = type(v)
    if is_class_type(t):
        module_name = t.__module__
        if BUILTIN_MODULE_NAME != module_name:
            return module_name
    return ""


def translate_special_mock_return_value(value):
    """
    Translate possible special values in a mock call return value.

    :param value:
    :return:
    """
    v = value
    if isinstance(value, ExpressionModel):
        special_value = eval_special_value(value.expression)
        if special_value is not None:
            v = special_value
    return v


def collect_return_values(mock_calls: List[MockCallModel]):
    """

    even if a method has no return value it implictly returns "None"
    "None" can also be a valid return value
    Without setting the mock return value in the generated code,
    a mock object will be returned.
    So evern if the return value is None, it is still necessary to set it.
    """
    return [translate_special_mock_return_value(m.return_value) for m in mock_calls]


def add_indentation(s: str, level: int) -> str:
    lines = s.split("\n")
    indented_lines = add_indentation_to_str_list(lines=lines, level=level)
    return "\n".join(indented_lines)


def add_indentation_to_str_list(lines: List[str], level: int) -> List[str]:
    leading_spaces = g_indent * level
    return [f"{leading_spaces}{line}" for line in lines]


def remove_empty(a_list: list) -> list:
    return [e for e in a_list if e]



from __future__ import annotations

import logging
import re
from typing import Any, List, Tuple, Optional

from testscribe import global_var
from testscribe.api.mock_api import throw
from testscribe.ignore import IGNORED
from testscribe.mock_proxy import MockProxy
from testscribe.model_type import (
    MockNameModel,
    MockModel,
    get_mock_by_name,
    ExpressionModel,
    SetModel,
    TestModel,
)
from testscribe.reflection_util import get_module_str
from testscribe.special_type import NoDefault
from testscribe.user_triggered_exception import UserTriggeredException
from testscribe.util import remove_trailing_numbers, generic_transform

logger = logging.getLogger(__name__)


def is_simple_value(v) -> bool:
    """
    Does the value consist of str, bool, int, float, None, MockProxy only
    including all elements if the value is a collection.

    MockProxy will be represented by its name in the generated code.
    Thus, it is considered a "simple" value that doesn't need special handling.

    :param v:
    :return:
    """
    if v is None:
        return True
    t = type(v)
    if t in (str, bool, int, float, MockProxy):
        return True
    if t in (set, tuple, list):
        items = v
    elif t is dict:
        items = v.values()
    else:
        return False

    for e in items:
        if not is_simple_value(e):
            return False
    return True


def eval_special_value(
    value_str: str,
):
    """
    Check if the expression evalutes to UserTriggeredException or ignored
    if so return it otherwise return None.
    It doesn't raise an exception.
    """
    global_dict = {throw.__name__: throw, "ignore": IGNORED}
    # noinspection PyBroadException
    try:
        value = eval(value_str, {}, global_dict)
        if isinstance(value, UserTriggeredException) or value is IGNORED:
            return value
        else:
            return None
    except Exception:
        return None


def transform_default_value(default: Any, t: type) -> str:
    """
    Translate the default value into a string suitable as raw input.

    :param default:
    :param t: the expected type. Note it can be Any
    """
    if default is NoDefault:
        return transform_no_default_value(t)
    else:
        return transform_real_default_value(default)


def transform_real_default_value(default: Any) -> str:
    if default is None:
        return "None"
    elif isinstance(default, str):
        return default
    elif isinstance(default, ExpressionModel):
        return default.expression
    elif isinstance(default, SetModel) or is_simple_value(default):
        return repr(default)
    else:
        return transform_mock_names_to_mock_expression(
            v=default,
            test_to_infer_default_inputs=global_var.g_test_to_infer_default_inputs,
        )


def transform_no_default_value(t: type) -> str:
    if t is bool:
        return "False"
    else:
        return ""


def transform_mock_names_to_mock_expression(
    v: Any, test_to_infer_default_inputs: Optional[TestModel]
) -> str:
    """
    It's used to transform mock object value in a default value back to an expression
    string which can be used to recreate the same mock object.
    """
    if test_to_infer_default_inputs:

        def transform_one_mock_name_to_mock_expression_with_test(val: Any):
            return transform_one_mock_name_to_mock_expression(
                value=val, test_to_infer_default_inputs=test_to_infer_default_inputs
            )

        transformed = generic_transform(
            v=v, transform_func=transform_one_mock_name_to_mock_expression_with_test
        )
    else:
        transformed = v
    return repr(transformed)


def transform_one_mock_name_to_mock_expression(
    value: Any, test_to_infer_default_inputs: TestModel
):
    if isinstance(value, MockNameModel):
        mock_name = value.name
        mock_model: MockModel = get_mock_by_name(
            mocks=test_to_infer_default_inputs.mocks, name=mock_name
        )
        # Translate back to what users will have to type to create the same mock object.
        base_mock_name = remove_trailing_numbers(mock_name)
        return ExpressionModel(f"m({mock_model.spec_str}, '{base_mock_name}')")
    else:
        return value


def get_string_value(value: str) -> str:
    """
    Handle strings that are quoted which are necessary to support multiline strings.

    :param value: raw string input
    :return: string value
    """
    if value.startswith('"') or value.startswith("'"):
        # Allow multiline string input by allowing e.g. "a\nb"
        # i.e. enclose the input with "" or ''
        # to use escape characters.
        return eval(value)
    elif value.startswith(r"\"") or value.startswith(r"\'") or value.startswith(r"\\"):
        # The downside is that users have to remember to escape the
        # leading " or ' with \" or \'
        # if they want to use the input string that starts with " or '
        # as it is without escape characters
        # To input a string that starts with \, escape it with \\
        return value[1:]
    else:
        return value


def import_modules_from_expression(user_input: str) -> Tuple[dict, List[str]]:
    """
    Import modules referenced in the user input expression

    :param user_input:
    :return: tuple ( dict ( top package name -> top package reference),
    list of full module names found )
    """
    full_names = get_possible_fully_qualified_names(user_input)
    imports = {}
    module_names = []
    for name in full_names:
        top_package, module_name = try_import_module(name)
        if top_package:
            imports[top_package.__name__] = top_package
            module_names.append(module_name)
    return imports, module_names


def get_possible_fully_qualified_names(expression: str) -> List[str]:
    """
    Given a python expression string, return a list of possible fully qualified
    class names in the string.
    :param expression:
    :return:
    """
    all_captures = QUALIFIED_CLASS_NAME_PATTERN.findall(expression)
    return [full for full, _ in all_captures]


QUALIFIED_CLASS_NAME_PATTERN = re.compile(r"((\w+\.)+\w+)")


def try_import_module(name: str):
    """

    :param name: the fully qualified name of an object
    :return: a tuple ( top level package, full module name string ) or ( None, None) if
    the module can't be loaded.
    """
    while name:
        module_name = get_module_str(name)
        if module_name:
            top_package = import_module(module_name)
            if top_package:
                return top_package, module_name
        # The next component may be a class name etc. not a module name
        # e.g. inspect.Parameter.empty inspect is the module not inspect.Parameter.
        # todo: detect errors that a module name may be mis-spelled.
        name = module_name
    return None, None


def import_module(module_name: str):
    try:
        top_package = __import__(module_name)
        return top_package
    except ModuleNotFoundError:
        # This warning may show up when generating a new test and the file contains
        # a test with incorrectly detected module name in an expression.
        # todo: re-evalue
        # log(f"Failed to import possible module: ({module_name}). " f"exception: {e}")
        return None



import logging
from pathlib import Path
from typing import Optional

from testscribe import global_var
from testscribe.config import Config
from testscribe.execution import execute_and_generate, run_setup_func
from testscribe.execution_util import init
from testscribe.file_info import get_module
from testscribe.input_params import create_input_params, get_scribe_file_path
from testscribe.load_scribe_file import load_or_create_model
from testscribe.log import log

logger = logging.getLogger(__name__)


def get_output_root_dir(output_root_dir: Optional[Path], config: Config) -> Path:
    if output_root_dir:
        log(f"Output root directory cmd line option: {output_root_dir}")
    else:
        # use the one in the config
        output_root_dir = config.output_root_path
    normalized_output_root_dir = output_root_dir.resolve()
    log(f"Output root directory: {normalized_output_root_dir}")
    return normalized_output_root_dir


def create_test(
    source_file: Path,
    function_name: str,
    output_root_dir: Optional[Path],
    ask_for_test_name: bool,
    ask_for_description: bool,
    config_file_path: Optional[Path],
):
    # The user defined setup function may call patch.
    # make sure the mode is set up before the setup function is called.
    global_var.g_test_generating_mode = True
    config = init(config_file_path=config_file_path)

    log(f"Source file to test: {source_file}")
    log(f"Function to test: {function_name}")
    module = get_module(source_file)
    log(f"Target module to test: {module.get_module_str()}")

    normalized_output_root_dir = get_output_root_dir(output_root_dir, config)
    input_params = create_input_params(
        module=module,
        function_name=function_name,
        output_root_dir=normalized_output_root_dir,
        ask_for_test_name=ask_for_test_name,
        ask_for_description=ask_for_description,
    )
    scribe_file_path = get_scribe_file_path(
        output_root_dir=normalized_output_root_dir, module=module
    )
    all_tests = load_or_create_model(
        file_path=scribe_file_path,
        full_module_name=input_params.module.get_module_str(),
    )
    run_setup_func(config.setup_function)
    execute_and_generate(
        input_params=input_params, all_tests=all_tests, index_of_test_to_update=-1
    )



from __future__ import annotations


class IgnoreReturnValue:
    def __repr__(self):
        # Use the string as the return value of a mock call.
        # It has to be a valid expression since it will be used in the generated test.
        return "'Ignored'"


# Use this variable when the special input expression "ignore" is used for a mock
# return value to ignore the return value.
# It's useful if the return value is not used in the production code.
IGNORED = IgnoreReturnValue()



from types import ModuleType, FunctionType
from typing import Callable, Any

from testscribe.complex_value_util import (
    contain_complex_value,
    object_contains_unstable_repr,
    has_custom_repr_method,
)
from testscribe.error import UnsupportedDataError
from testscribe.log import log
from testscribe.mock_proxy import is_mock_proxy
from testscribe.model_type import (
    ObjectModel,
    CallableModel,
    MockNameModel,
    ExpressionModel,
    ModuleModel,
    SetModel,
)
from testscribe.reflection_util import get_full_spec_name, get_module_str_from_object
from testscribe.type_util import (
    is_class_type,
)
from testscribe.util import generic_transform
from testscribe.value_util import InputValue


def transform_value(v):
    """
    Replace certain values that can't be encoded correctly in Yaml to special
    types.
    """
    return generic_transform(v=v, transform_func=transform_individual_value_to_model)


def transform_individual_value_to_model(v):
    t = type(v)
    if is_mock_proxy(v):
        # Only mock names are needed here since this is a reference.
        # The mock object details are encoded elsewhere.
        return MockNameModel(v.name_test_scribe_)
    elif isinstance(v, InputValue):
        # Only the expression part is needed to generate
        # unit tests. And the generated YAML file will be more compact
        # and readable.
        return ExpressionModel(v.expression)
    elif callable(v) and hasattr(v, "__qualname__"):
        # A class instance may be callable by implementing the __call__ method.
        # but since it is not a class, it doesn't have the __qulname__ attribute
        # e.g. an instance of MockCall
        # Only the name is needed.
        return create_callable_model(v)
    elif t == ModuleType:
        # Modules can't be loaded from its native yaml representation correctly.
        # repr of a module is like
        # <module 'test_data.simple' from '.../code/test-scribe/python/tests/test_data/simple.py'>
        return ModuleModel(name=v.__name__)
    elif t == set:
        # A native set YAML representation is not stable. The order of the elements may
        # change when it is regenerated.
        # todo: handle the case when the set can't be sorted.
        if contain_complex_value(v):
            # There is currently no reliable way to identify the items in a set
            # to generate individual asserts.
            log(
                f"The set:\n{repr(v)}\n"
                "contains complex objects. It is currently not supported."
            )
            raise UnsupportedDataError(
                "Sets that contain complex objects are not supported."
            )
        return SetModel(elements=sorted(v))
    elif is_class_type(t):
        # pyyaml need to load the modules in order to load the objects.
        # Also, it is not clear how pyyaml can encode all objects using
        # constructors.
        # Loading arbitray Python objects requires the use of pyyaml
        # unsafe_loader.
        # After execution, there is no need to reconstruct the objects anyway.
        return transform_class(v)
    else:
        return v


def create_callable_model(v: Callable):
    if v == FunctionType:
        # the "function" type is callable. And the name "function" is undefined.
        # translate to the proper name using the "types" module.
        name = "FunctionType"
        module = "types"
    else:
        name = v.__qualname__
        module = get_module_str_from_object(v)
    return CallableModel(name=name, module=module)


def transform_class(value):
    t = type(value)
    assert is_class_type(t)
    full_class_name = get_full_spec_name(t)
    if can_use_repr(value):
        # Use __repr__ instead of __str__ to capture more details of the object
        # Ideally __repr__ can be used to reconstruct the exact object.
        # see https://dzone.com/articles/python-str-vs-repr
        representation = repr(value)
        members = {}
    else:
        representation = ""
        # vars function includes only properties and its values.
        # It doesn't include methods and builtin properties.
        # vars() only works when the value has __dict__ attribute
        # module values for example don't have such an attribute.
        # such values should have been filtered out before calling this method.
        # In case such a value is missed, set the members to {} just in case.
        members = (
            {name: transform_value(v) for name, v in vars(value).items()}
            if hasattr(value, "__dict__")
            else {}
        )

    return ObjectModel(type=full_class_name, repr=representation, members=members)


def can_use_repr(value: Any) -> bool:
    return has_custom_repr_method(type(value)) and not object_contains_unstable_repr(
        value
    )



from __future__ import annotations

from inspect import Parameter
from typing import Any, List

from click import echo, prompt

from testscribe.api.io_provider import IOProvider
from testscribe.constant import GET_TEST_NAME_PROMPT
from testscribe.context import Context
from testscribe.namedvalues import NamedValues
from testscribe.parameter_value_input_cli import get_parameter_value_cli
from testscribe.value_input_cli import get_one_value_cli
from testscribe.value_input_util import get_string_value


class CLI(IOProvider):
    def log(self, s: str) -> None:
        echo(s)

    def get_parameter_value(
        self, param_info_list: List[Parameter], defaults: list, context: Context
    ) -> NamedValues:
        return get_parameter_value_cli(
            param_info_list=param_info_list, defaults=defaults
        )

    def get_one_value(
        self, prompt_name: str, name: str, t: type, default: Any, context: Context
    ) -> Any:
        return get_one_value_cli(prompt_name=prompt_name, t=t, default=default)

    def get_test_description(self, default: str) -> str:
        return get_test_description_cli(default)

    def get_short_test_name(self, default_short_name: str) -> str:
        return get_short_test_name_cli(default_short_name)


def get_test_description_cli(default: str):
    prompt_str = "Provide a description of the test."
    input_str = prompt(prompt_str, default=default)
    # todo format the text so that each line is within a maximum size.
    return get_string_value(input_str)


def get_short_test_name_cli(default_short_name: str) -> str:
    return prompt(GET_TEST_NAME_PROMPT, default=default_short_name)



from typing import List

from testscribe.config import g_indent
from testscribe.model_type import PatchModel


def generate_patch_str(patches: List[PatchModel]) -> str:
    if not patches:
        return ""
    patch_str_list = [f"patch('{p.target}', {repr(p.replacement)})" for p in patches]
    patch_str = ", ".join(patch_str_list)
    return f"\n{g_indent}with {patch_str}:"



import logging
from typing import List

from testscribe.code_gen_util import join_lines
from testscribe.gather_referenced_modules import (
    gather_import_statements_for_referenced_modules,
)
from testscribe.model_type import (
    TestModel,
    AllTests,
)

logger = logging.getLogger(__name__)


def need_mock_support(tests: List[TestModel]) -> bool:
    for t in tests:
        if t.mocks:
            return True
    return False


def gather_import_statements_for_mock_support(tests: List[TestModel]) -> List[str]:
    if need_mock_support(tests):
        # todo: only import when there is call verifications
        return [
            "from testscribe.api.mock_api import get_normalized_mock_calls",
            "from unittest.mock import ANY, call, create_autospec",
        ]
    else:
        return []


def gather_import_statement_for_patch_support(tests: List[TestModel]) -> str:
    for t in tests:
        if t.patches:
            return "from unittest.mock import patch"
    return ""


def need_exception_support(tests: List[TestModel]) -> bool:
    for t in tests:
        if t.exception:
            return True
    return False


def gather_import_statement_for_exception_support(tests: List[TestModel]) -> str:
    if need_exception_support(tests):
        return "import pytest"
    else:
        return ""


def get_target_for_a_test(t: TestModel):
    target_class_name = t.target_class_name
    if target_class_name:
        # todo: handle nested classes
        target_name = target_class_name
    else:
        target_name = t.target_func_name
    return target_name


def gather_import_statements_for_test_targets(all_tests: AllTests) -> str:
    targets = {get_target_for_a_test(t) for t in all_tests.tests}
    target_str = ", ".join(sorted(targets))
    return f"from {all_tests.module} import {target_str}"


def generate_import_statement_str(all_tests: AllTests) -> str:
    statements = []
    tests = all_tests.tests
    statements.extend(gather_import_statements_for_referenced_modules(tests))
    statements.extend(gather_import_statements_for_mock_support(tests))
    statements.append(gather_import_statement_for_patch_support(tests))
    statements.append(gather_import_statement_for_exception_support(tests))
    statements.append(gather_import_statements_for_test_targets(all_tests))
    result = join_lines(statements, prepend_new_line=False) + "\n"
    return result



from __future__ import annotations

INVALID_TEST_INDEX = -1
GET_TEST_NAME_PROMPT = (
    "Test name help: 'test_' prefix will be added automatically. "
    "Use a leading '_' to include the target function name as "
    "part of the prefix.\n"
    "Test name:"
)



import logging
from pathlib import Path
from typing import Any, Union, Callable

from typer import prompt

from testscribe import global_var
from testscribe.config import Config
from testscribe.constant import INVALID_TEST_INDEX
from testscribe.execution import execute_and_generate, run_setup_func
from testscribe.execution_util import ERROR_RETURN_CODE, init
from testscribe.input_params import create_input_params
from testscribe.load_scribe_file import load_scribe_file
from testscribe.log import log
from testscribe.mock_proxy import MockProxy
from testscribe.mocking_support import (
    patch_with_mock_internal,
    patch_with_expression_internal,
)
from testscribe.model_type import (
    AllTests,
    TestModel,
    PatchModel,
    MockNameModel,
    ExpressionModel,
    get_mock_by_name,
)
from testscribe.module import get_module_from_str, Module
from testscribe.reflection_util import get_symbol
from testscribe.util import remove_trailing_numbers

logger = logging.getLogger(__name__)


def update_test_cmd(
    scribe_file_path: Path,
    test_name: str,
) -> int:
    config = init()
    log(f"Update the test {test_name} in {scribe_file_path}")
    all_tests = load_scribe_file(scribe_file_path)
    index = all_tests.get_test_index_by_name(test_name)
    logger.debug(f"Index of the test being updated: {index}")
    if index != INVALID_TEST_INDEX:
        update_test(
            index=index,
            all_tests=all_tests,
            scribe_file_path=scribe_file_path,
            config=config,
        )
        return 0
    else:
        log(
            f"The test ({test_name}) doesn't exist in the file "
            f"({scribe_file_path})."
        )
        return ERROR_RETURN_CODE


def update_test(
    index: int, all_tests: AllTests, scribe_file_path: Path, config: Config
):
    test = all_tests.tests[index]

    # Save to global variables to avoid passing this information via
    # many layers of calls which don't need this information.
    global_var.g_index_of_test_to_update = index
    module_name = all_tests.module
    log(f"Updating module ({module_name}).")
    module = get_module_from_str(module_name)
    output_root_dir = infer_output_root_dir_from_module(module, scribe_file_path)
    log(f"Output root directory: {output_root_dir}")
    create_patches_from_existing_test(all_tests.tests[index])
    # Patches defined in a setup function overrides the existing patches
    # thus they have to be called later.
    # todo: add a test
    run_setup_func(config.setup_function)
    # Unlike the create command, update cmd indicates that the scribe file
    # is likely to be maintained. In such a case, a meaningful name and
    # description become more useful.
    # It's likely that the test name and descriptions may need updating.
    input_params = create_input_params(
        module=module,
        function_name=test.target_func_name,
        output_root_dir=output_root_dir,
        ask_for_test_name=True,
        ask_for_description=True,
    )
    execute_and_generate(
        input_params=input_params, all_tests=all_tests, index_of_test_to_update=index
    )


def infer_output_root_dir_from_module(module: Module, scribe_file_path: Path) -> Path:
    """

    :param module: the module the scribe file is targeting.
    :param scribe_file_path:
    :return: the output root path
    """
    package_depth = len(module.get_package_name_list())
    return scribe_file_path.parents[package_depth]


def create_patches_from_existing_test(test: TestModel):
    if test.patches:
        log("Recreating patches from the existing test.")
        for p in test.patches:
            create_patcher_from_model(patch_model=p, test=test)


def create_patcher_from_model(patch_model: PatchModel, test: TestModel) -> bool:
    if should_recreate_patch(patch_model):
        really_create_patcher_from_model(patch_model=patch_model, test=test)
        return True
    else:
        log(f"Skipped the patch for {patch_model.target}.")
        return False


def should_recreate_patch(patch_model: PatchModel) -> bool:
    prompt_str = (
        f"Create patch with target ( {patch_model.target} )"
        f" replacement ( {patch_model.replacement} )?"
    )
    # noinspection PyTypeChecker
    # the default value should be a bool here not a string
    return prompt(prompt_str, default=True, type=bool)


def really_create_patcher_from_model(patch_model: PatchModel, test: TestModel) -> None:
    value = patch_model.replacement
    if isinstance(value, MockNameModel):
        create_mock_patcher_from_model(
            mock_name=value.name, target=patch_model.target, test=test
        )
    else:
        patch_with_expression_internal(
            target_str=patch_model.target, expression=get_expression_str(value)
        )


def create_mock_patcher_from_model(
    mock_name: str, target: Union[str, type, Callable, MockProxy], test: TestModel
):
    mock = get_mock_by_name(mocks=test.mocks, name=mock_name)
    spec = get_symbol(mock.spec_str)
    normalized_name = remove_trailing_numbers(mock_name)
    patch_with_mock_internal(target=target, mock_name=normalized_name, spec=spec)


def get_expression_str(value: Any) -> str:
    if isinstance(value, ExpressionModel):
        return value.expression
    return repr(value)



import logging
from dataclasses import dataclass
from typing import Any, List, Optional, Dict

from testscribe.constant import INVALID_TEST_INDEX
from testscribe.namedvalues import NamedValues
from testscribe.reflection_util import get_module_str
from testscribe.util import BUILTIN_MODULE_NAME, consistent_hash_str

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class ObjectModel:
    type: str
    repr: str
    members: dict

    # The default __hash__ function can't hash a field that is a dictionary.
    # which is mutable. The error will be "TypeError: unhashable type: 'dict'".
    def __hash__(self):
        return consistent_hash_str(str(self))

    def __str__(self):
        if self.members:
            value_str = f"members ({self.members})"
        else:
            value_str = f"repr ({self.repr})"
        return f"Object(type ({get_type_str(self)}), {value_str})"

    def __repr__(self):
        if self.repr:
            module_str = get_module_str(self.type)
            prefix = module_str + "." if module_str else ""
            # Use the fully qualified name for the constructor assuming
            # the repr from a custom type starts with the type name.
            return prefix + self.repr
        else:
            # fall back to the str representation.
            # this type of object can't be used as keys in a dictionary.
            return str(self)


@dataclass(frozen=True)
class CallableModel:
    name: str
    module: str

    def __repr__(self):
        """
        Override the default implementation to allow the class instance to be
        used correctly when generating the test code.
        It can appear as a parameter value.
        :return:
        """
        if self.module == BUILTIN_MODULE_NAME:
            return self.name
        else:
            return f"{self.module}.{self.name}"


@dataclass(frozen=True)
class MockNameModel:
    name: str

    def __repr__(self):
        """
        Override the default implementation to allow the class instance to be
        used correctly when generating the test code.
        :return:
        """
        return self.name

    def __hash__(self):
        return consistent_hash_str(self.name)

    def __str__(self):
        return f"Mock( name: {self.name} )"


@dataclass(frozen=True)
class ExpressionModel:
    expression: str

    def __repr__(self):
        """
        Override the default implementation to allow the class instance to be
        used correctly when generating the test code.
        :return:
        """
        return self.expression

    def __str__(self):
        return f"Expression( {self.expression} )"


@dataclass(frozen=True)
class ModuleModel:
    name: str

    def __str__(self):
        return f"Module( {self.name} )"


@dataclass(frozen=True)
class SetModel:
    elements: list

    def __repr__(self):
        """
        Override the default implementation to allow the instance to be
        used correctly when generating the test code.
        It can appear as a parameter value.
        :return:
        """
        return f"set({repr(self.elements)})"


@dataclass
class MockCallModel:
    name: str
    parameters: NamedValues
    return_value: Any


@dataclass
class MockModel:
    name: str
    spec_str: str
    calls: List[MockCallModel]
    attributes: dict


@dataclass
class PatchModel:
    target: str
    replacement: Any


@dataclass
class ExceptionModel:
    type: str
    message: str

    def __str__(self):
        return f"Exception: type ( {self.type} ), message ( {self.message} )"


@dataclass
class TestModel:
    # The name field can be derived from the short_name field easily.
    # however, it is needed as a unique key in some cases.
    # It's thus kept for convenience.
    name: str
    # The name that doesn't include the test_<optiontal function name> prefix
    # It makes it easier to update the test name after the target function name
    # is changed and the function name is used in the test name prefix.
    # The name is required since it is better not to have to derive it from
    # the name attribute.
    short_name: str
    description: str
    target_func_name: str
    target_class_name: str
    init_parameters: NamedValues
    # parameters to the function/method. If there is an exception in the constructor
    # this value will be None
    parameters: Optional[NamedValues]
    exception: Optional[ExceptionModel]
    result: Any

    # Favor list over dictionary because the order of the list is stable.
    # When a new test is added, don't want to introduce unnecssary changes
    # to existing tests.
    mocks: List[MockModel]
    patches: List[PatchModel]


def get_mock_by_name(mocks: List[MockModel], name: str) -> Optional[MockModel]:
    for m in mocks:
        if m.name == name:
            return m
    return None


def sort_tests(tests: List[TestModel]) -> List[TestModel]:
    # sort tests based on the target
    # It's expected that most of the time, it is beneficial to have
    # these tests to be close to each other.
    return sorted(tests, key=get_test_sort_key)


def get_test_sort_key(test: TestModel) -> str:
    return test.target_class_name + test.target_func_name


def build_name_to_index_dict(tests: List[TestModel]) -> dict:
    return {t.name: index for index, t in enumerate(tests)}


@dataclass
class AllTests:
    tests: List[TestModel]
    module: str
    name_to_index: Dict[str, int]

    def __init__(self, module: str, tests: List[TestModel]):
        self.module = module
        self.tests = sort_tests(tests)
        self.name_to_index = build_name_to_index_dict(self.tests)
        logger.debug(f"Test name to index dict: {self.name_to_index}")

    def get_test_index_by_name(self, test_name: str) -> int:
        index = self.name_to_index.get(test_name, INVALID_TEST_INDEX)
        assert index == INVALID_TEST_INDEX or test_name == self.tests[index].name
        return index

    def does_test_exist(self, test_name: str) -> bool:
        return self.get_test_index_by_name(test_name) != INVALID_TEST_INDEX


def add_test(all_tests: AllTests, test: TestModel) -> AllTests:
    # Return a new AllTest to allow the constructor to maintain
    # the integrety of the internal cache and sorted properties.
    logger.info(f"Add test: {test.name}")
    # Add new tests before old tests
    # This behavior is assumed by the logic of guessing the default value
    # to use the latest test targeting the same function.
    new_tests = [test] + all_tests.tests
    return AllTests(module=all_tests.module, tests=new_tests)


def update_test(all_tests: AllTests, index: int, test: TestModel) -> AllTests:
    # make a copy of the list to avoid modifying the existing list
    new_tests = list(all_tests.tests)
    new_tests[index] = test
    return AllTests(module=all_tests.module, tests=new_tests)


def delete_test_by_name(all_tests: AllTests, name: str) -> AllTests:
    new_tests = [t for t in all_tests.tests if t.name != name]
    return AllTests(module=all_tests.module, tests=new_tests)


MODEL_TYPE_TO_STR: Dict[type, str] = {
    # More user friendly/readable type names.
    CallableModel: "Callable",
    MockNameModel: "Mock",
    ExpressionModel: "Expression",
    ModuleModel: "Module",
    SetModel: "Set",
    ExceptionModel: "Exception",
}


def get_type_str(v: Any) -> str:
    if isinstance(v, ObjectModel):
        t = v.type
    else:
        t = type(v)
    return MODEL_TYPE_TO_STR.get(t, str(t))



"""
The in memory representation of all the tests
"""
from __future__ import annotations

import logging
from typing import Optional, List

from testscribe import global_var
from testscribe.description import get_test_description
from testscribe.input_params import (
    InputParams,
    get_target_function_name,
    get_target_class_name,
)
from testscribe.mock_proxy import MockProxy
from testscribe.model_type import (
    MockCallModel,
    MockModel,
    PatchModel,
    ExceptionModel,
    TestModel,
    AllTests,
    add_test,
    update_test,
)
from testscribe.patcher import Patcher
from testscribe.reflection_util import get_full_spec_name
from testscribe.runtime_values import RuntimeValues
from testscribe.test_name import get_test_name

logger = logging.getLogger(__name__)


def generate_model(
    input_params: InputParams,
    runtime_values: RuntimeValues,
    all_tests: AllTests,
    index_of_test_to_update: int,
) -> AllTests:
    target_func_name = get_target_function_name(input_params)
    short_name, test_name = get_test_name(
        all_tests=all_tests,
        index_of_test_to_update=index_of_test_to_update,
        ask_for_test_name=input_params.ask_for_test_name,
        target_func_name=target_func_name,
    )
    description = get_test_description(
        all_tests=all_tests,
        index_of_test_to_update=index_of_test_to_update,
        ask_for_description=input_params.ask_for_description,
    )
    new_test = TestModel(
        name=test_name,
        short_name=short_name,
        description=description,
        target_func_name=target_func_name,
        target_class_name=get_target_class_name(input_params),
        init_parameters=runtime_values.init_values,
        parameters=runtime_values.values,
        exception=get_exception_model(runtime_values.exception),
        result=runtime_values.result,
        mocks=get_mock_models(global_var.g_name_mock_dict.values()),
        patches=get_patch_models(global_var.g_patchers.values()),
    )
    return create_new_all_tests(
        all_tests=all_tests,
        index_of_test_to_update=index_of_test_to_update,
        new_test=new_test,
    )


def get_exception_model(exception: Optional[Exception]) -> Optional[ExceptionModel]:
    if exception:
        return ExceptionModel(
            type=get_full_spec_name(type(exception)), message=str(exception)
        )
    else:
        return None


def get_mock_models(mocks: List[MockProxy]) -> List[MockModel]:
    return [generate_mock_model(m) for m in mocks]


def generate_mock_model(mock: MockProxy):
    call_models = [
        MockCallModel(
            name=c.method_name,
            parameters=c.args,
            return_value=c.return_value,
        )
        for c in mock.calls_test_scribe_
        # todo: c.args is None only when MockCall is not called.
        # can this happen when a mock's method is referenced but not called?
        # document an example.
        if c.args
    ]
    return MockModel(
        name=mock.name_test_scribe_,
        spec_str=get_full_spec_name(mock.spec_test_scribe_),
        calls=call_models,
        attributes=mock.attributes_transformed_test_scribe_,
    )


def get_patch_models(patches: List[Patcher]) -> List[PatchModel]:
    return [create_patch_model(p) for p in patches]


def create_patch_model(p: Patcher):
    # This assumes the replacement_spec contains transformed value.
    return PatchModel(target=p.target, replacement=p.replacement_spec)


def create_new_all_tests(
    all_tests: AllTests, index_of_test_to_update: int, new_test: TestModel
):
    if index_of_test_to_update < 0:
        new_all_tests = add_test(all_tests=all_tests, test=new_test)
    else:
        new_all_tests = update_test(
            all_tests=all_tests, index=index_of_test_to_update, test=new_test
        )
    return new_all_tests



from __future__ import annotations

from typing import Dict, List

from testscribe import scribe_file_key as key
from testscribe.model_type import (
    AllTests,
    TestModel,
    PatchModel,
    MockModel,
    MockCallModel,
)
from testscribe.namedvalues import NamedValues


def encode_model(all_tests: AllTests) -> dict:
    # It's possible to store the tests as dictionary keyed by the
    # test name.
    # however, since before python 3.6 (2016/12)
    # the dictionary won't keep the key insertion order.
    # Storing as a dictionary may result in unnecessary changes.
    # Also, the order is harder to control when updating, creating new tests.
    # see
    # https://stackoverflow.com/questions/1867861/how-to-keep-keys-values-in-same-order-as-declared
    # https://en.wikipedia.org/wiki/History_of_Python
    tests = encode_tests(all_tests)
    content = {
        key.FORMAT_VERSION: 1,
        key.MODULE: all_tests.module,
        key.TESTS: tests,
    }
    return content


def encode_tests(all_tests: AllTests) -> List[dict]:
    return [encode_one_test(t) for t in all_tests.tests]


def encode_one_test(test_model: TestModel) -> Dict:
    test = encode_name_and_description(test_model)
    # Encode information most tests share first
    # ensure consistency in the order of information among tests.
    encode_target(test=test, test_model=test_model)
    encode_parameters(test=test, test_model=test_model)
    encode_result(test=test, test_model=test_model)
    # encode patches before mocks since mocks may be more complex
    encode_patches(test=test, patches=test_model.patches)
    encode_mocks(test=test, mocks=test_model.mocks)
    return test


def encode_name_and_description(test_model: TestModel) -> dict:
    test = {key.NAME: test_model.name, key.SHORT_NAME: test_model.short_name}
    description = test_model.description
    if description:
        test[key.DESCRIPTION] = description
    return test


def encode_target(test: dict, test_model: TestModel) -> dict:
    target = {key.NAME: test_model.target_func_name}
    if test_model.target_class_name:
        target[key.CLASS_NAME] = test_model.target_class_name
    test[key.TARGET] = target
    return test


def encode_parameters(test: dict, test_model: TestModel) -> dict:
    init_values = encode_named_values(test_model.init_parameters)
    if init_values:
        # Don't include the key since a function (not a method)  will not have a need
        # for init parameters.
        test[key.INIT_PARAMETERS] = init_values
    if test_model.parameters:
        test[key.PARAMETERS] = encode_named_values(test_model.parameters)
    return test


def encode_named_values(values: NamedValues) -> List[dict]:
    # Encode as a list of dictionaries with the key "name" and "value"
    # Can't be a dictionary since some values may not have names.
    return [
        {key.NAME: name, key.VALUE: value} if name else {key.VALUE: value}
        for name, value in values.as_list()
    ]


def encode_result(test: dict, test_model: TestModel) -> dict:
    exception_model = test_model.exception
    if exception_model:
        test[key.EXCEPTION] = {
            key.TYPE: exception_model.type,
            key.MESSAGE: exception_model.message,
        }
    else:
        test[key.RESULT] = test_model.result
    return test


def encode_patches(test: dict, patches: List[PatchModel]) -> dict:
    if patches:
        test[key.PATCHES] = [
            {key.TARGET: p.target, key.PATCH_REPLACEMENT: p.replacement}
            for p in patches
        ]
    return test


def encode_mocks(test: dict, mocks: List[MockModel]) -> Dict:
    if mocks:
        test[key.MOCKS] = [encode_a_mock(m) for m in mocks]
    return test


def encode_a_mock(m: MockModel) -> Dict:
    encoded = {
        key.NAME: m.name,
        key.SPEC: m.spec_str,
    }
    calls = encode_mock_calls(m)
    if calls:
        encoded[key.CALLS] = calls
    attributes = m.attributes
    if attributes:
        encoded[key.ATTRIBUTES] = attributes
    return encoded


def encode_mock_calls(m: MockModel) -> List[dict]:
    return [encode_one_mock_call(c) for c in m.calls]


def encode_one_mock_call(mock_call: MockCallModel) -> dict:
    name = mock_call.name
    encoded = encode_one_mock_call_name(name)
    # noinspection PyTypeChecker
    encoded[key.PARAMETERS] = encode_named_values(mock_call.parameters)
    return_value = mock_call.return_value
    if return_value is not None:
        encoded[key.RETURN_VALUE] = return_value
    return encoded


def encode_one_mock_call_name(name: str) -> dict:
    return {key.NAME: name} if name else {}



import typing
from functools import partial
from inspect import Parameter, Signature, getmembers, signature, getattr_static
from typing import Any, Callable, get_type_hints, List, Iterable

from testscribe.custom_type import Spec
from testscribe.error import Error
from testscribe.special_type import SingleLineString
from testscribe.type_util import (
    is_class_type,
    is_typing_callable_type,
    is_function_instance,
    get_type_args,
)
from testscribe.util import BUILTIN_MODULE_NAME, load_object

TYPE_TO_TYPE_NAME = {
    Any: "any",
    SingleLineString: "str",
    Parameter.empty: "unknown",
    Signature.empty: "unknown",
}


def get_type_name(t: type) -> str:
    if t in TYPE_TO_TYPE_NAME:
        return TYPE_TO_TYPE_NAME[t]
    if is_class_type(t):
        return get_full_spec_name(t)
    if hasattr(t, "__name__"):
        return t.__name__
    return str(t)


def get_full_spec_name(t: Spec) -> str:
    """

    :param t:
    :return: full name including module
    """
    module_name = t.__module__
    module_str = "" if module_name == BUILTIN_MODULE_NAME else module_name + "."
    if hasattr(t, "__qualname__"):
        return f"{module_str}{t.__qualname__}"
    else:
        # Callable or Callable[[int], int] doesn't have the __qualname__
        # attribute.
        return repr(t)


def has_param_names(spec: Spec):
    """
    return True if the spec has information about the names of the parameters.
    e.g. Callable[[int], int] doesn't.
    :param spec:
    :return:
    """
    # both class and functions have this attribute. Callable doesn't.
    return hasattr(spec, "__qualname__")


def get_typing_callable_return_type(t: type):
    assert is_typing_callable_type(t)
    args = get_type_args(t)
    if args:
        return args[1]
    else:
        return Parameter.empty


def is_method(clazz: type, name: str):
    """
    Return true if the attribute is a method defined by this class

    :param clazz: the Class definition
    :param name:
    :return:
    """
    if name in dir(clazz):
        attrib = getattr(clazz, name)
        return callable(attrib)
    return False


def get_method(clazz: type, method_name: str):
    methods = getmembers(clazz)
    for n, func in methods:
        if n == method_name and callable(func):
            # Can't use is_function_instance(func) to valify that the member is a method
            # __str__ method of a mock object for example has a class type.
            return func
    return None


def is_instance_method(clazz: type, method_name: str):
    try:
        m = getattr_static(clazz, method_name)
        return not (isinstance(m, staticmethod) or isinstance(m, classmethod))
    except AttributeError:
        return False


def get_return_type(func: Callable) -> type:
    sig = signature(func)
    return_annotation = sig.return_annotation
    if isinstance(return_annotation, str):
        # delayed annotation processing is enabled.
        # see https://www.python.org/dev/peps/pep-0563/
        name_to_annotation_dict = get_type_hints(func)
        return name_to_annotation_dict["return"]
    else:
        return return_annotation


def get_param_list(func: Callable) -> List[Parameter]:
    """
    Return a normalized list of Parameter
    This is needed because with from __future__ import annotations
    the types returned from signature() will all be strings.
    get_type_hints() however will process the type strings into type classes.

    :param func: a function or method object
    :return:
    """
    sig = signature(func)
    # Unlike signature(), get_type_hint() doesn't automatically return
    # the annotation for the __init__ method.
    func_to_get_type_info = (
        func if is_function_instance(func) else getattr(func, "__init__")
    )
    name_to_annotation_dict = get_type_hints(func_to_get_type_info)
    param_info_list = sig.parameters.values()

    def normalize(p: Parameter) -> Parameter:
        annotation = p.annotation
        if isinstance(annotation, str):
            name = p.name
            annotation_type = name_to_annotation_dict[name]
            return Parameter(name=name, kind=p.kind, annotation=annotation_type)
        else:
            return p

    return [normalize(p) for p in param_info_list]


def get_method_signature_for_caller(clazz: type, name: str):
    """
    Get the method signature as used by the caller.
    e.g. if this is an instance method, the signature should not include self.
    :param clazz:
    :param name:
    :return:
    """
    method = get_method(clazz, name)
    if method is None:
        raise Error(f"{name} is not a method of the class {clazz}")
    # todo: Some builtin objects don't provide signatures. e.g. dict.items()
    if is_instance_method(clazz, name):
        return signature(partial(method, None))
    else:
        return signature(method)


def get_bound_arguments(sig, args: Iterable, kwargs: dict) -> dict:
    return sig.bind(*args, **kwargs).arguments


def remove_brackets(full_name: str) -> str:
    """
    Remove everything after the first "["
    for the input like typing.Callable[[typing.Any], bool]
    """
    name, _, _ = full_name.partition("[")
    return name


def get_module_and_symbol(full_name: str) -> typing.Tuple[str, str]:
    """
    Can't handle a method name e.g. test_data.simple.C.bar

    :param full_name: assumes the last component as separated by "." is the symbol name.
    :return: tuple( module name, symbol name)
    """
    name = remove_brackets(full_name)
    components = name.split(".")
    module_parts = components[:-1]
    return ".".join(module_parts), components[-1]


def get_module_str(full_name: str) -> str:
    return get_module_and_symbol(full_name)[0]


def get_module_str_from_object(obj: Any) -> str:
    # Objects not implemented in python may not have the __module__ attribute
    if hasattr(obj, "__module__"):
        return obj.__module__
    else:
        return ""


def get_symbol(full_name: str):
    module_name, symbol = get_module_and_symbol(full_name)
    obj = load_object(symbol_name=symbol, module_str=module_name)
    if obj is None:
        raise Error(f"{full_name} is not a valid identifier.")
    return obj



from __future__ import annotations

import logging
from typing import Tuple, Set

from testscribe import global_var
from testscribe.model_type import AllTests
from testscribe.util import convert_camel_case_to_snake_case

logger = logging.getLogger(__name__)


def get_test_name(
    all_tests: AllTests,
    index_of_test_to_update: int,
    ask_for_test_name: bool,
    target_func_name: str,
) -> Tuple[str, str]:
    short_name = get_short_name(
        all_tests=all_tests,
        ask_for_test_name=ask_for_test_name,
        index_of_test_to_update=index_of_test_to_update,
    )
    test_name = create_test_name(
        all_tests=all_tests,
        index_of_test_to_update=index_of_test_to_update,
        short_name=short_name,
        target_func_name=target_func_name,
    )
    return short_name, test_name


def get_short_name(
    all_tests: AllTests, ask_for_test_name: bool, index_of_test_to_update: int
) -> str:
    default_short_name = get_default_short_name(
        all_tests=all_tests, index_of_test_to_update=index_of_test_to_update
    )
    raw_short_name = get_raw_short_name(
        default_short_name=default_short_name, ask_for_test_name=ask_for_test_name
    )
    short_name = create_proper_short_name(raw_short_name)
    return short_name


def get_default_short_name(all_tests: AllTests, index_of_test_to_update: int) -> str:
    if index_of_test_to_update < 0:
        return "_"
    else:
        return all_tests.tests[index_of_test_to_update].short_name


def get_raw_short_name(default_short_name: str, ask_for_test_name: bool) -> str:
    if ask_for_test_name:
        return global_var.g_io.get_short_test_name(default_short_name)
    else:
        return default_short_name


def create_proper_short_name(raw_short_name: str) -> str:
    return convert_camel_case_to_snake_case(raw_short_name.replace(" ", "_"))


def create_test_name(
    all_tests: AllTests,
    index_of_test_to_update: int,
    short_name: str,
    target_func_name: str,
) -> str:
    base_name = create_proper_test_name(
        short_name=short_name, target_func_name=target_func_name
    )
    test_name = generate_unique_test_name(
        base_name=base_name,
        all_tests=all_tests,
        test_to_update_index=index_of_test_to_update,
    )
    return test_name


def create_proper_test_name(short_name: str, target_func_name: str) -> str:
    # It's easier to read and simpler to implement if we don't include
    # class name in the test name.
    # Assuming it is rare there will be methods/functions with same name in
    # the same module. Even when it happens, the test names will still be unique
    # in the same file.
    name = replace_leading_underscore_with_function_name(
        s=short_name, function_name=target_func_name
    )
    return "test_" + name


def replace_leading_underscore_with_function_name(s: str, function_name: str) -> str:
    if s == "_":
        return function_name
    if s.startswith("_"):
        return function_name + s
    return s


def generate_unique_test_name(
    base_name: str, all_tests: AllTests, test_to_update_index: int
) -> str:
    name = base_name
    count = 0
    other_test_names = get_other_test_names(
        test_to_update_index=test_to_update_index, all_tests=all_tests
    )
    while name in other_test_names:
        # This will try to find the smallest postfix number that avoids name conflict.
        # e.g. If there is a test_foo and test_foo_2 test, it will generate test_foo_1
        count += 1
        name = f"{base_name}_{count}"
    return name


def get_other_test_names(test_to_update_index: int, all_tests: AllTests) -> Set[str]:
    tests = all_tests.tests
    return {
        # If this is part of an update test operation, the name of this test
        # should be ignored for the purpose of avoiding name conflicts
        # since it will be overwritten.
        tests[i].name
        for i in range(len(tests))
        if i != test_to_update_index
    }



import logging.config
import traceback
from dataclasses import dataclass
from inspect import signature
from pathlib import Path
from typing import Optional, Callable, List, Any, Dict

from testscribe import global_var
from testscribe.context import Context
from testscribe.error import UnsupportedDataError
from testscribe.execution_util import (
    infer_unit_test_file_path_from_scribe_file,
    call_target_function,
)
from testscribe.generate_tests import generate_unit_test_file
from testscribe.input_params import (
    InputParams,
    get_target_function_name,
    get_target_class_name,
    get_scribe_file_path,
)
from testscribe.log import log
from testscribe.model import generate_model
from testscribe.model_type import TestModel, AllTests, get_type_str
from testscribe.namedvalues import NamedValues
from testscribe.parameter_value_input import get_parameter_value
from testscribe.patcher import Patcher
from testscribe.reflection_util import get_full_spec_name
from testscribe.runtime_values import RuntimeValues
from testscribe.save_scribe_file import generate_scribe_file
from testscribe.transformer import transform_value
from testscribe.value_util import get_value

logger = logging.getLogger(__name__)


@dataclass
class CallResult:
    arguments: Optional[NamedValues]
    result: Any
    exception: Optional[Exception]


def transform_named_values(named_values: NamedValues) -> NamedValues:
    return NamedValues(transform_value(named_values.as_list()))


def get_args_and_call(func: Callable, default: NamedValues) -> CallResult:
    """
    Gather arguments for the given function and call it.

    :param func:
    :param default: the default for the arguments
    :return:
    """
    name = func.__qualname__
    log(f"Getting parameters for the function ({name})")
    description = f"Parameters for the function ({name})"
    named_values = get_parameter_value(
        func=func, context=Context(description=description), default=default
    )
    # Have to make a copy before invoking the function
    transformed_named_values = transform_named_values(named_values)
    values = [get_value(value) for _, value in named_values.as_list()]
    sig = signature(func)
    args = sig.bind(*values)
    arg_str = named_values.as_arg_str()
    log(f"Calling {name}({arg_str})")
    exception = None
    try:
        ret_value = call_target_function(func, args)
    except Exception as e:
        log(f"The function {func.__qualname__} throws an exception.")
        # todo: filter out the system frames at the bottom of the stack.
        log(traceback.format_exc())
        ret_value = None
        exception = e
    return CallResult(
        arguments=transformed_named_values, result=ret_value, exception=exception
    )


def get_default_init_parameters(
    test_to_infer_default_inputs: Optional[TestModel],
) -> NamedValues:
    if test_to_infer_default_inputs:
        return test_to_infer_default_inputs.init_parameters
    else:
        return NamedValues()


def create_instance(
    clazz: type, test_to_infer_default_inputs: Optional[TestModel]
) -> CallResult:
    """
    Create an instance of the class being tested.
    :param test_to_infer_default_inputs:
    :param clazz:
    :return:
    """
    name = clazz.__qualname__
    log(f"Prepare to create an instance of the class: {name}")
    default = get_default_init_parameters(test_to_infer_default_inputs)
    return get_args_and_call(func=clazz, default=default)


def get_test_to_infer_default_inputs(
    tests: List[TestModel],
    index_of_test_to_update: int,
    function_name: str,
    class_name: str,
) -> Optional[TestModel]:
    if index_of_test_to_update >= 0:
        logger.info(
            f"Use the test being updated to retrive default values."
            f" index: {index_of_test_to_update}, name: {tests[index_of_test_to_update].name}"
        )
        return tests[index_of_test_to_update]
    # use the newest test with the same target function to infer default inputs
    # Note that this assumes newer tests appear in this list earlier.
    for t in tests:
        if t.target_func_name == function_name and t.target_class_name == class_name:
            logger.info(
                f"Use a test with the same target to retrive default. name: {t.name}"
            )
            return t
    logger.info("Can't find a test with the same target to retrive defaults values.")
    return None


def execute_and_generate(
    input_params: InputParams,
    all_tests: AllTests,
    index_of_test_to_update: int,
):
    runtime_values = create_instance_and_run(
        all_tests=all_tests,
        index_of_test_to_update=index_of_test_to_update,
        input_params=input_params,
    )
    generate_test(
        input_params=input_params,
        runtime_values=runtime_values,
        all_tests=all_tests,
        index_of_test_to_update=index_of_test_to_update,
    )


def create_instance_and_run(
    all_tests: AllTests, index_of_test_to_update: int, input_params: InputParams
) -> RuntimeValues:
    test_to_infer_default_inputs = get_test_to_infer_default_inputs(
        tests=all_tests.tests,
        index_of_test_to_update=index_of_test_to_update,
        function_name=get_target_function_name(input_params),
        class_name=get_target_class_name(input_params),
    )
    global_var.g_test_to_infer_default_inputs = test_to_infer_default_inputs
    try:
        function_instance = get_function_instance(
            input_params=input_params,
            test_to_infer_default_inputs=test_to_infer_default_inputs,
        )
        call_result = run_target_function(
            constructor_exception=function_instance.exception,
            func=function_instance.func,
            test_to_infer_default_inputs=test_to_infer_default_inputs,
        )
    finally:
        stop_patches(global_var.g_patchers)

    return RuntimeValues(
        init_values=function_instance.init_values,
        instance=function_instance.target_instance,
        values=call_result.arguments,
        result=transform_and_show_result(call_result),
        exception=call_result.exception,
    )


def transform_and_show_result(call_result: CallResult) -> Any:
    if call_result.exception:
        return None
    raw_result = call_result.result
    try:
        transformed_result = transform_value(raw_result)
    except UnsupportedDataError as e:
        log(f"Result:\n{raw_result}")
        raise e
    # show the transformed result to have better display of certain types
    # e.g. showing the members of an object.
    show_result(transformed_result)
    return transformed_result


def show_result(result: Any) -> str:
    result_str = f"***** Result:\n{show_result_internal(result)}\n***** Result end"
    log(result_str)
    return result_str


def show_result_internal(result: Any) -> str:
    # todo: force the str method when a complex object
    # such as ObjectModel object is used inside a container
    # to have a better display.
    return f"type: {get_type_str(result)}\nvalue:\n{result}"


def stop_patches(patchers: Dict[str, Patcher]) -> None:
    for p in patchers.values():
        p.instance.stop()


def run_target_function(
    constructor_exception: Optional[Exception],
    func: Callable,
    test_to_infer_default_inputs: Optional[TestModel],
) -> CallResult:
    if constructor_exception:
        return CallResult(arguments=None, result=None, exception=constructor_exception)
    else:
        log("Prepare to call the target function.")
        default = (
            test_to_infer_default_inputs.parameters
            if test_to_infer_default_inputs
            else NamedValues()
        )
        logger.debug(f"run target function. Default parameters: {default}")
        return get_args_and_call(func=func, default=default)


@dataclass
class FunctionInstance:
    func: Callable
    # Exception thrown when creating a class instance if the func is a method
    exception: Optional[Exception]
    # If the func is a method, the bounded class instance. Otherwise None
    target_instance: Any
    # If the func is a method, the arguments to the bounded class' constructor
    init_values: NamedValues


def get_function_instance(
    input_params: InputParams, test_to_infer_default_inputs: TestModel
) -> FunctionInstance:
    exception = None
    clazz = input_params.clazz
    func = input_params.func
    if clazz:
        # todo: handle class methods. They don't need to create a class
        #  instance.
        call_result = create_instance(clazz, test_to_infer_default_inputs)
        target_instance = call_result.result
        init_values = call_result.arguments
        exception = call_result.exception
        if not exception:
            function_name = get_target_function_name(input_params)
            func = getattr(target_instance, function_name)
    else:
        target_instance = None
        init_values = NamedValues()
    return FunctionInstance(
        func=func,
        init_values=init_values,
        target_instance=target_instance,
        exception=exception,
    )


def generate_test(
    input_params: InputParams,
    runtime_values: RuntimeValues,
    all_tests: AllTests,
    index_of_test_to_update: int,
):
    new_all_tests = generate_model(
        input_params=input_params,
        runtime_values=runtime_values,
        all_tests=all_tests,
        index_of_test_to_update=index_of_test_to_update,
    )
    scribe_file_path = get_scribe_file_path(
        output_root_dir=input_params.output_root_dir, module=input_params.module
    )
    save_file(scribe_file_path=scribe_file_path, all_tests=new_all_tests)
    end(scribe_file_path=scribe_file_path)


def end(scribe_file_path: Path):
    global_var.g_io.end(scribe_file_path)


def save_file(scribe_file_path: Path, all_tests: AllTests):
    generate_scribe_file(scribe_file_path=scribe_file_path, all_tests=all_tests)
    test_file_path = infer_unit_test_file_path_from_scribe_file(scribe_file_path)
    generate_unit_test_file(test_file_path=test_file_path, all_tests=all_tests)


def run_setup_func(f: Optional[Callable]):
    if f is not None:
        log(f"Calling the setup function {get_full_spec_name(f)}.")
        f()



from yaml import add_representer, add_constructor, Dumper, Loader

from testscribe.model_type import (
    ObjectModel,
    MockNameModel,
    ExpressionModel,
    CallableModel,
    ModuleModel,
    SetModel,
)

# Use representer functions instead of a subclass of YAMLObject
# to represent objects to avoid translating the classes.

MOCK_NAME_YAML_TAG = "!testscribe/mock"


def mock_name_model_representer(dumper, mock_name_model: MockNameModel):
    # This method has the added benefit of generating a one line representation
    # When using a YAMLObject based approach, the result looks like
    # !testscribe/mock 'm_mock_call_model'
    return dumper.represent_scalar(MOCK_NAME_YAML_TAG, mock_name_model.name)


def mock_name_model_constructor(loader, node):
    name = loader.construct_scalar(node)
    return MockNameModel(name=name)


CALLABLE_YAML_TAG = "!testscribe/callable"
CALLABLE_YAML_KEY_NAME = "name"
CALLABLE_YAML_KEY_MODULE = "module"


def callable_model_representer(dumper, callable_model: CallableModel):
    value = {
        CALLABLE_YAML_KEY_NAME: callable_model.name,
        CALLABLE_YAML_KEY_MODULE: callable_model.module,
    }
    return dumper.represent_mapping(CALLABLE_YAML_TAG, value)


def callable_model_constructor(loader, node):
    value = loader.construct_mapping(node)
    return CallableModel(
        name=value[CALLABLE_YAML_KEY_NAME], module=value[CALLABLE_YAML_KEY_MODULE]
    )


EXPRESSION_YAML_TAG = "!testscribe/expression"


def expression_model_representer(dumper, expression_model: ExpressionModel):
    return dumper.represent_scalar(EXPRESSION_YAML_TAG, expression_model.expression)


def expression_model_constructor(loader, node):
    expression = loader.construct_scalar(node)
    return ExpressionModel(expression=expression)


OBJECT_YAML_TAG = "!testscribe/object"
OBJECT_YAML_KEY_TYPE = "type"
OBJECT_YAML_KEY_REPR = "repr"
OBJECT_YAML_KEY_MEMBERS = "members"


def object_model_representer(dumper, object_model: ObjectModel):
    """
    Use representer functions instead of a subclass of YAMLObject
    to represent ObjectModel object to skip fields that are empty.

    :param dumper:
    :param object_model:
    :return:
    """
    value = {OBJECT_YAML_KEY_TYPE: object_model.type}
    if object_model.repr:
        value[OBJECT_YAML_KEY_REPR] = object_model.repr
    else:
        value[OBJECT_YAML_KEY_MEMBERS] = object_model.members
    return dumper.represent_mapping(OBJECT_YAML_TAG, value)


def object_model_constructor(loader, node):
    value = loader.construct_mapping(node)
    t = value[OBJECT_YAML_KEY_TYPE]
    represantation = value.get(OBJECT_YAML_KEY_REPR, "")
    members = value.get(OBJECT_YAML_KEY_MEMBERS, {})
    return ObjectModel(type=t, repr=represantation, members=members)


MODULE_YAML_TAG = "!testscribe/module"


def module_model_representer(dumper, module_model: ModuleModel):
    return dumper.represent_scalar(MODULE_YAML_TAG, module_model.name)


def module_model_constructor(loader, node):
    name = loader.construct_scalar(node)
    return ModuleModel(name=name)


SET_YAML_TAG = "!testscribe/set"


def set_model_representer(dumper: Dumper, set_model: SetModel):
    return dumper.represent_sequence(SET_YAML_TAG, set_model.elements)


def set_model_constructor(loader: Loader, node):
    elements = loader.construct_sequence(node)
    return SetModel(elements=elements)


def add_representer_for_custom_tags():
    add_representer(ObjectModel, object_model_representer)
    add_representer(MockNameModel, mock_name_model_representer)
    add_representer(CallableModel, callable_model_representer)
    add_representer(ExpressionModel, expression_model_representer)
    add_representer(ModuleModel, module_model_representer)
    add_representer(SetModel, set_model_representer)


def add_constructor_for_custom_tags():
    add_constructor(OBJECT_YAML_TAG, object_model_constructor)
    add_constructor(MOCK_NAME_YAML_TAG, mock_name_model_constructor)
    add_constructor(EXPRESSION_YAML_TAG, expression_model_constructor)
    add_constructor(CALLABLE_YAML_TAG, callable_model_constructor)
    add_constructor(MODULE_YAML_TAG, module_model_constructor)
    add_constructor(SET_YAML_TAG, set_model_constructor)



from __future__ import annotations

import logging
from collections.abc import Iterable
from inspect import signature, Signature
from typing import Any, List, Optional

from testscribe import global_var
from testscribe.context import Context
from testscribe.custom_type import Spec
from testscribe.execution_util import show_user_call_stack
from testscribe.log import log
from testscribe.model_type import TestModel, MockCallModel, get_mock_by_name
from testscribe.namedvalues import NamedValues
from testscribe.reflection_util import (
    has_param_names,
    get_typing_callable_return_type,
    get_method,
    get_return_type,
    is_instance_method,
)
from testscribe.special_type import NoDefault
from testscribe.type_util import is_typing_callable_type, is_function_instance
from testscribe.user_triggered_exception import UserTriggeredException
from testscribe.value_util import get_value

logger = logging.getLogger(__name__)

STR_MAGIC_METHOD_NAME = "__str__"


class MockCall:
    # Define class variables here to provide type information for the instant variables
    # spec field is missing here because there is no good default value
    # to provide. If set to str for example, the field spec will be detected
    # as a method since str is callable.
    method_name = ""
    mock_name = ""
    previous_call_count = 0
    args = NamedValues()

    def __init__(
        self,
        method_name: str,
        mock_name: str,
        spec: Spec,
        previous_call_count: int,
    ):
        """

        :param method_name: '' if this is a direct call on the mock object
        :param mock_name: the name of the mock object associated with this call
        :param spec: the spec of the mock object
        :param previous_call_count: how many times this method of the mock object
        has been called before this call
        """
        #
        self.method_name = method_name
        self.mock_name = mock_name
        self.spec = spec
        self.previous_call_count = previous_call_count

        # The args is a transformed deep copy of the mock call parameters.
        self.args = NamedValues()
        self.return_value = None

    def __call__(self, *args, **kwargs) -> Any:
        return self.call_internal(*args, **kwargs)

    def call_internal(self, *args, **kwargs) -> Any:
        self.args = save_args(self, args, kwargs)
        value = get_return_value(self)

        # Unlike mock attributes, the return value is not used again
        # during the execution of the target function. So there is
        # no need to keep a separate copy of the transformed value.
        self.return_value = prepare_value_for_storage(value)
        return get_true_return_value(value)


def save_args(mock_call: MockCall, args: Iterable, kwargs: dict) -> NamedValues:
    args_list = get_args_list(mock_call=mock_call, args=args, kwargs=kwargs)

    # transform the value early to keep a copy of the original value
    # in case the value is changed after execution.
    transformed = prepare_value_for_storage(args_list)
    return NamedValues(transformed)


def prepare_value_for_storage(v: Any):
    # avoid circular reference
    from testscribe.transformer import transform_value

    return transform_value(v)


def get_args_list(mock_call: MockCall, args: Iterable, kwargs: dict) -> list:
    # todo: handle the TypeError exception that may be thrown in bind
    #  if the parameters don't match the signature
    # https://docs.python.org/3/library/inspect.html#inspect.Signature
    sig = get_mock_call_signature(mock_call)
    return get_args_list_internal(
        method_name=mock_call.method_name,
        sig=sig,
        spec=mock_call.spec,
        args=args,
        kwargs=kwargs,
    )


def get_mock_call_signature(mock_call: MockCall) -> Signature:
    # todo: Some builtin objects don't provide signatures. e.g. dict.items()
    spec = mock_call.spec
    method_name = mock_call.method_name
    if method_name:
        method = get_method(spec, method_name)
        sig = signature(method)
    else:
        # This is a call on the mock directly.
        # Either the mock is a function or it is a call to instantiate
        # a class.
        sig = signature(spec)
    logger.debug(f"Mock call signature {sig}")
    return sig


def get_args_list_internal(
    method_name: str,
    sig: Signature,
    spec: Spec,
    args: Iterable,
    kwargs: dict,
):
    if is_instance_method(clazz=spec, method_name=method_name):
        # Since this is a method call, the first argument is always self.
        bound_args = sig.bind(None, *args, **kwargs)
        args_list = list(bound_args.arguments.items())[1:]
    else:
        args_list = get_func_args_list(sig=sig, spec=spec, args=args, kwargs=kwargs)
    return args_list


def get_func_args_list(
    sig: Signature, spec: Spec, args: Iterable, kwargs: dict
) -> list:
    if has_param_names(spec):
        bound_args = sig.bind(*args, **kwargs)
        # noinspection PyTypeChecker
        args_list = list(bound_args.arguments.items())
    else:
        positional_args = [("", v) for v in args]
        args_list = positional_args + list(kwargs.items())
    return args_list


def get_return_value(mock_call: MockCall) -> Any:
    description = get_call_description(mock_call)
    default = get_default_mock_call_return_value(mock_call)
    return_type = get_mock_call_return_type(
        method_name=mock_call.method_name, spec=mock_call.spec
    )
    # avoid circular references
    from testscribe.value_input import get_one_value

    value = get_one_value(
        prompt_name="the return value",
        t=return_type,
        context=Context(description),
        default=default,
    )
    log(f"Mock call return value: {repr(value)}")
    return value


def get_call_description(mock_call: MockCall) -> str:
    arg_str_display = get_arg_str_display(mock_call)
    subject = get_call_subject(mock_call)
    description = f"{subject} is called{arg_str_display}."
    log(description)
    show_user_call_stack(additional_num_system_frame=4)
    return description


def get_arg_str_display(mock_call: MockCall):
    arg_str = mock_call.args.as_arg_str()
    arg_str_display = f"\nwith: {arg_str}" if arg_str else ""
    return arg_str_display


def get_call_subject(mock_call: MockCall) -> str:
    mock_name = mock_call.mock_name
    method_name = mock_call.method_name
    if method_name:
        return f"{mock_name}'s {mock_call.method_name} method"
    else:
        return mock_name


def get_default_mock_call_return_value(mock_call: MockCall) -> Any:
    if mock_call.method_name == STR_MAGIC_METHOD_NAME:
        return f"mock {mock_call.mock_name}"
    return get_default_return_value_internal(
        mock_call=mock_call,
        test_to_infer_default_inputs=global_var.g_test_to_infer_default_inputs,
    )


def get_default_return_value_internal(
    mock_call: MockCall, test_to_infer_default_inputs: Optional[TestModel]
):
    # disable the debug log since it makes generated tests more complex.
    # logger.debug(f"infer default input from test: {test_to_infer_default_inputs}")
    if test_to_infer_default_inputs is None:
        return NoDefault
    existing_mock = get_mock_by_name(
        mocks=test_to_infer_default_inputs.mocks, name=mock_call.mock_name
    )
    if existing_mock is None:
        return NoDefault
    return get_default_return_value_from_models(
        mock_call_models=existing_mock.calls, mock_call=mock_call
    )


def get_default_return_value_from_models(
    mock_call_models: List[MockCallModel], mock_call: MockCall
) -> Any:
    existing_return_values = [
        call.return_value
        for call in mock_call_models
        if call.name == mock_call.method_name
    ]
    size = len(existing_return_values)
    # There will be cases when the default value is not picked
    # correctly.
    # todo: make all possible values selectable.
    if size > 0:
        # The new production code may have different number of calls to
        # this method. Select from the existing test's calls in a round-robin fashion.
        return existing_return_values[mock_call.previous_call_count % size]
    else:
        return NoDefault


def get_mock_call_return_type(method_name: str, spec: Spec) -> type:
    if method_name == STR_MAGIC_METHOD_NAME:
        # __str__'s signature returned by the get_return_type is _empty
        return_type = str
    elif method_name:
        method = get_method(spec, method_name)
        return_type = get_return_type(func=method)
    elif is_function_instance(spec):
        return_type = get_return_type(spec)
    elif is_typing_callable_type(spec):
        return_type = get_typing_callable_return_type(spec)
    else:
        # This is a call to instantiate a class
        # the return type should be the target class but the function
        # signature has None as return type
        # todo: support Callable e.g. Callable, Callable[[int], str]
        return_type = spec
    logger.debug(f"return type {return_type}")
    return return_type


def get_true_return_value(value: Any) -> Any:
    """
    Return the return value needed to run the target function

    :param value:
    :return:
    """
    true_value = get_value(value)
    if isinstance(true_value, UserTriggeredException):
        user_exeception = true_value.exception
        raise user_exeception
    else:
        return true_value



from __future__ import annotations

from typing import Callable

from testscribe import global_var
from testscribe.context import Context
from testscribe.namedvalues import NamedValues, NameNotFound
from testscribe.reflection_util import get_param_list
from testscribe.special_type import NoDefault


def get_parameter_value(
    func: Callable, context: Context, default: NamedValues
) -> NamedValues:
    param_info_list = get_param_list(func=func)
    if not param_info_list:
        return NamedValues()
    default_values = [
        get_default_value_from_old_params(default=default, index=index, name=param.name)
        for index, param in enumerate(param_info_list)
    ]
    # g_io is initialized after the module is imported
    # so directly importing g_io will get the default value only.
    return global_var.g_io.get_parameter_value(
        param_info_list=param_info_list, defaults=default_values, context=context
    )


def get_default_value_from_old_params(default: NamedValues, index: int, name: str):
    default_length = default.get_size()
    default_value = default.get_value_by_name(name)
    if default_value is NameNotFound:
        # If the name can't be found, it may be because the parameter
        # name has changed.
        # fall back to default to the parameter in the same position.
        if index < default_length:
            _, default_value = default[index]
        else:
            default_value = NoDefault
    return default_value



from __future__ import annotations

import logging
import re
from inspect import Signature, Parameter
from typing import Any

import testscribe.api.mock_api
from testscribe import global_var
from testscribe.api.alias import g_aliases
from testscribe.complex_value_util import generic_is_special_type
from testscribe.error import InputError
from testscribe.log import log
from testscribe.mock_proxy import MockProxy, is_mock_proxy
from testscribe.reflection_util import get_full_spec_name
from testscribe.type_util import get_class_type, get_type_origin, get_type_args
from testscribe.value_input_util import is_simple_value, import_modules_from_expression
from testscribe.value_util import InputValue

logger = logging.getLogger(__name__)


def eval_expression(user_input: str, t: type) -> Any:
    expression = expand_aliases(t=t, user_input=user_input)
    value = eval_with_injected_names(expression)
    # Have to use the user input string to generate the scribe and unit test
    # files if the input contains
    # classes or non builtin functions.
    # Can't use whether a fully qualified module name is detected as a test
    # because the m(type) expression should not be sufficient to be
    # stored as an expression.
    post_value = process_mock_marker(t=t, v=value)
    return wrap_input_value(expression=expression, v=post_value)


def expand_aliases(t: type, user_input: str) -> str:
    expression = expand_string_aliases(user_input)
    return expand_class_alias(t=t, expression=expression)


def expand_string_aliases(expression: str):
    for alias, full_str in g_aliases.items():
        expression = expand_one_string_alias(
            expression=expression, alias=alias, full_str=full_str
        )
    return expression


def expand_one_string_alias(expression: str, alias: str, full_str: str) -> str:
    """
    Replace user defined string aliases with their replacement strings in the
    expression.

    :param expression: input expression
    :param alias:
    :param full_str: replacement string
    :return: expanded expression
    """
    # The alias has to be surrounded by at least one non word or digit
    # Use ?: non-capturing group to avoid replacing these parts.
    # see
    # https://stackoverflow.com/questions/3512471/what-is-a-non-capturing-group-in-regular-expressions
    pattern = f"(?P<pre>^|\\W+){alias}(?P<after>$|\\W+)"
    logger.debug(f"Alias pattern: {pattern}")
    # Add the captured groups one before and one after the alias to avoid changing them.
    # Use named group instead of \1 and \2 to allow full_str to start with numbers.
    replacement_pattern = r"\g<pre>" + full_str + r"\g<after>"
    logger.debug(f"expression: {expression}")
    replaced, replacement_count = re.subn(
        pattern=pattern, repl=replacement_pattern, string=expression
    )
    if replacement_count:
        log(f"Expanded alias: {alias} {replacement_count} times.")
        log(f"Result after the expantion: {replaced}")
    return replaced


def expand_class_alias(t: type, expression: str) -> str:
    """
    Replace the expression with a new expression with the 'c' alias replaced
    with the class constructor.

    :param t: the expected type.
    :param expression: input expression
    :return: expanded expression if it contains a valid 'c' alias that can be matched
    to a class type
    """
    class_type = get_class_type(t)
    if class_type:
        # allows the name c to be used as an alias for the class type name.
        if expression.startswith("c("):
            expression = get_full_spec_name(class_type) + expression[1:]
    return expression


def eval_with_injected_names(expression: str) -> Any:
    # Mock object names are defined inside the generated functions, so they have
    # local scope.
    global_dict = global_var.g_name_mock_dict.copy()
    # Make the m function available to the expression.
    # The 'm' mock name is reserved so that there will be no name conflict.
    global_dict["m"] = testscribe.api.mock_api.m
    top_level_modules, _ = import_modules_from_expression(expression)
    value = eval(expression, top_level_modules, global_dict)
    return value


def process_mock_marker(t: type, v: Any) -> Any:
    """
    Translate the 'm' (mock marker) into mock objects in expressions
    e.g.
    [m] for a type List[C] is translated to [MockProxy(C)]
    See the unit tests for more examples.

    :param t: the expected type for the value
    :param v: the value the expression evaluates to
    :return: the translated value
    """
    if is_type_infor_unavailable(t):
        return v
    if is_mock_marker(v):
        return process_mock_marker_value(t=t)
    elif isinstance(v, list):
        return process_mock_marker_list(t=t, v=v)
    elif isinstance(v, tuple):
        return process_mock_marker_tuple(t=t, v=v)
    # mock marker in a set is not supported since a match by position
    # in the value is impossible.
    elif isinstance(v, dict):
        return process_mock_marker_dict(t=t, v=v)
    else:
        return v


def is_type_infor_unavailable(t: type) -> bool:
    return t in (Any, Signature.empty, Parameter.empty)


def is_mock_marker(value) -> bool:
    return value is testscribe.api.mock_api.m


def process_mock_marker_value(t: type):
    class_type = get_class_type(t)
    if class_type:
        return MockProxy(spec=class_type)
    else:
        raise_mock_marker_type_mismatch_error(f"The type ({t}) can't be mocked.")


def raise_mock_marker_type_mismatch_error(msg: str):
    raise InputError(msg)


def process_mock_marker_list(t: type, v: Any):
    # todo: handle Iterable type in a similar way
    if get_type_origin(t) is list:
        item_types = get_type_args(t)
        if len(item_types) == 1:
            # if List is used, there may not be any type information for its items
            (item_type,) = item_types
            return [process_mock_marker(v=item, t=item_type) for item in v]
    # The type may be Any or unknown, skip processing the mock marker.
    # In that case, it should be fine when the input value doesn't contain any
    # mock marker.
    # todo: check if the type is indeed Any or unknown, otherwise throw an exception.
    return v


def process_mock_marker_tuple(t: type, v: Any):
    if get_type_origin(t) is tuple:
        item_types = get_normaized_item_types(t=t, v=v)
        if len(v) != len(item_types):
            raise InputError(
                f"tuple value ({v}) size doesn't match the tuple type ({t})."
            )
        return tuple(
            [
                process_mock_marker(v=item, t=item_type)
                for item, item_type in zip(v, item_types)
            ]
        )
    else:
        # The type may be Any or unknown, skip processing the mock marker.
        # In that case, it should be fine when the input value doesn't contain any
        # mock marker.
        # todo: check if the type is indeed Any or unknown, otherwise throw an exception.
        return v


def get_normaized_item_types(t: type, v: Any):
    item_types = get_type_args(t)
    if len(item_types) == 2 and item_types[1] == Ellipsis:
        item_types_list = []
        for i in range(len(v)):
            item_types_list.append(item_types[0])
        item_types = tuple(item_types_list)
    return item_types


def process_mock_marker_dict(t: type, v: Any):
    if get_type_origin(t) is dict:
        args = get_type_args(t)
        if len(args) == 2:
            _, value_type = args
            return {
                key: process_mock_marker(t=value_type, v=value)
                for key, value in v.items()
            }
    # The dict type doesn't provide enough type information to support
    # the mock marker.
    # The type may be Any or unknown, skip processing the mock marker.
    # In that case, it should be fine when the input value doesn't contain any
    # mock marker.
    # todo: check if the type is indeed Any or unknown, otherwise throw an exception.
    return v


def wrap_input_value(expression: str, v: Any) -> Any:
    """
    Unlike output values which only need to be evaluated,
    input values need to be recreated in the generated tests.
    It's more reliable to use the user input instead of translating the value
    back to a proper constructor.
    e.g.
    If the expression evaluates to a class instance, it's easier to inspect/assert
    the class memembers than choosing the right constructor.
    Thus, is_simple_value only returns True for cases when constructing the value
    is reliable.
    :param expression: user input string after expanding aliases and mock markers.
    :param v: the value the expression evaluates to
    :return:
    """
    if is_simple_value(v):
        return v
    else:
        return process_complex_value(value=v, expression=expression)


def process_complex_value(value: Any, expression: str):
    if contain_mock_proxy(value) and is_m_function_in_expression(expression):
        # e.g. (m(test_data.simple.C), test_data.simple.C(1))
        # Generated code for parameter in the above example will be
        # (m_c, Object(type (test_data.simple.C), members ({'a': 1})))
        # It is ok if the MockProxy object is introduced via a
        # mock object name e.g. a parameter that is mocked.
        log(
            "Warning: mixing the 'm' function to create a mock object and other "
            "non primitive types in a user input will generate incorrect test files.\n"
            "Use a wrapper function to introduce the mock object via a named parameter "
            "and use that name in expressions instead.\n"
            "It still works to verify the test result one time and discard the "
            "generated test."
        )
        return value
        # todo: format the expression.
    return InputValue(expression=expression, value=value)


def contain_mock_proxy(value: Any) -> bool:
    """
    This method can't detect all cases. e.g.
    if a mockproxy object is used in the constructor of another object such as
    NamedValues.
    :param value:
    :return:
    """
    return generic_is_special_type(value=value, is_special_type_func=is_mock_proxy)


def is_m_function_in_expression(expression: str) -> bool:
    return M_FUNCTION_IN_EXPRESSION_PATTERN.search(expression) is not None


M_FUNCTION_IN_EXPRESSION_PATTERN = re.compile(r"(^|\W+)m *\(")



from __future__ import annotations

import logging
from pathlib import Path
from string import Template
from typing import Optional

from testscribe.code_gen_util import add_indentation
from testscribe.execution_util import remove_file_if_no_test
from testscribe.gen_mock_code import generate_mocks_str
from testscribe.gen_mock_verification_code import (
    generate_mock_call_verification_str,
)
from testscribe.gen_patch_code import generate_patch_str
from testscribe.generate_imports import generate_import_statement_str
from testscribe.log import log
from testscribe.model_type import (
    ExceptionModel,
    TestModel,
    AllTests,
)
from testscribe.namedvalues import NamedValues
from testscribe.result_assertion import generate_result_assertion

logger = logging.getLogger(__name__)


def generate_unit_test_file(test_file_path: Path, all_tests: AllTests):
    if remove_file_if_no_test(file_path=test_file_path, tests=all_tests.tests):
        return
    generated_test_str = generate_tests_output_string(all_tests=all_tests)
    # log(f"Generated test:\n{generated_test_str}")
    with test_file_path.open(mode="w") as f:
        f.write(generated_test_str)
    log(f"Wrote the generated test file to: {test_file_path}")


def generate_tests_output_string(all_tests: AllTests) -> str:
    import_statement_str = generate_import_statement_str(all_tests=all_tests)
    test_functions = [generate_one_test_function(t) for t in all_tests.tests]
    output_str = "\n\n".join([import_statement_str] + test_functions)
    return output_str


def generate_one_test_function(test: TestModel) -> str:
    template_str = """\
def $test_name():$docstring$mock_str$invocation$result_assertion$mock_call_verification
"""
    template = Template(template_str)
    docstring = generate_docstring(test.description)
    mocks = test.mocks
    mock_str = generate_mocks_str(mocks)
    mock_call_verification_str = generate_mock_call_verification_str(mocks)
    invocation = gen_invocation_str_with_patch(test)
    result_assertion = generate_result_assertion_str(test)
    output_str = template.substitute(
        test_name=test.name,
        docstring=docstring,
        func_name=test.target_func_name,
        mock_str=mock_str,
        invocation=invocation,
        result_assertion=result_assertion,
        mock_call_verification=mock_call_verification_str,
    )
    return output_str


def generate_docstring(description: str) -> str:
    if not description:
        return ""
    docstring_raw = f'"""\n{description}\n"""'
    docstring = add_indentation(docstring_raw, 1)
    return "\n" + docstring


def gen_invocation_str_with_patch(test: TestModel) -> str:
    patch_str = generate_patch_str(test.patches)
    create_and_invocation_str = gen_invocation_str_with_exception(test)
    indent_level = get_invocation_indent_level(patch_str)
    indented_invocation_str = add_indentation(create_and_invocation_str, indent_level)
    return f"{patch_str}\n{indented_invocation_str}"


def gen_invocation_str_with_exception(test: TestModel) -> str:
    inner_statement = gen_create_invocation_str(test)
    return wrap_exception_assertion(
        exception_model=test.exception, inner_statement=inner_statement
    )


def gen_create_invocation_str(test: TestModel) -> str:
    create_instance_statement = generate_create_instance_statement(
        target_class_name=test.target_class_name,
        init_parameters=test.init_parameters,
    )
    invocation_statement = generate_invocation_statement(test)
    return create_instance_statement + invocation_statement


def generate_create_instance_statement(
    target_class_name: str, init_parameters: NamedValues
) -> str:
    if target_class_name == "":
        return ""
    param_str = init_parameters.as_arg_str()
    return f"instance = {target_class_name}({param_str})\n"


def generate_invocation_statement(test: TestModel) -> str:
    parameters = test.parameters
    if parameters is None:
        # there is an exception in the constructor.
        return ""
    param_str = parameters.as_arg_str()
    instance_str = generate_target_instance_str(test.target_class_name)
    result_str = generate_result_assignment_str(does_test_have_exception(test))
    return f"{result_str}{instance_str}{test.target_func_name}({param_str})"


def does_test_have_exception(test: TestModel) -> bool:
    return test.exception is not None


def generate_result_assignment_str(has_exception: bool) -> str:
    result_str = "" if has_exception else "result = "
    return result_str


def generate_target_instance_str(target_class_name: str) -> str:
    instance_str = "instance." if target_class_name else ""
    return instance_str


def wrap_exception_assertion(
    exception_model: Optional[ExceptionModel], inner_statement: str
) -> str:
    if exception_model is None:
        return inner_statement
    template_str = """\
with pytest.raises($type) as exception_info:
$inner_statement
assert $message == str(exception_info.value)"""
    template = Template(template_str)
    output_str = template.substitute(
        type=exception_model.type,
        inner_statement=add_indentation(inner_statement, 1),
        message=repr(exception_model.message),
    )
    return output_str


def get_invocation_indent_level(patch_str: str) -> int:
    indent_level = 2 if patch_str else 1
    return indent_level


def generate_result_assertion_str(test: TestModel) -> str:
    result_assertion = "" if test.exception else generate_result_assertion(test.result)
    return result_assertion



import logging
from pathlib import Path

from testscribe.execution import save_file
from testscribe.execution_util import init
from testscribe.load_scribe_file import load_scribe_file
from testscribe.log import log
from testscribe.model_type import delete_test_by_name, AllTests

logger = logging.getLogger(__name__)


def delete_test_internal(scribe_file_path: Path, test_name: str, all_tests: AllTests):
    if all_tests.does_test_exist(test_name):
        new_all_tests = delete_test_by_name(all_tests=all_tests, name=test_name)
        save_file(scribe_file_path=scribe_file_path, all_tests=new_all_tests)
    else:
        log(f"The test name ({test_name}) doesn't exist")
        return


def delete_test(scribe_file_path: Path, test_name: str):
    init()
    log(f"Deleting the test {test_name} from {scribe_file_path}.")
    all_tests = load_scribe_file(scribe_file_path)
    # todo: if test_name is not specified, show a list of test names
    #  and allow users to choose one to delete
    delete_test_internal(
        scribe_file_path=scribe_file_path, test_name=test_name, all_tests=all_tests
    )



import logging
from pathlib import Path
from typing import List, Tuple, Optional, Any

from yaml import full_load

from testscribe import scribe_file_key as key
from testscribe.log import log
from testscribe.model_type import (
    MockCallModel,
    MockModel,
    PatchModel,
    ExceptionModel,
    TestModel,
    AllTests,
)
from testscribe.namedvalues import NamedValues
from testscribe.yaml_tag import add_constructor_for_custom_tags

logger = logging.getLogger(__name__)


def load_or_create_model(file_path: Path, full_module_name: str):
    if file_path.exists():
        logger.info(f"Load exsiting scribe file {file_path}.")
        return load_scribe_file(file_path)
    else:
        log(f"There is no existing test scribe file at {file_path}")
        return AllTests(module=full_module_name, tests=[])


def load_scribe_file(scribe_file: Path) -> AllTests:
    logger.debug(f"Loading the scribe file from: {scribe_file}")
    with scribe_file.open(mode="r") as f:
        content = load_from_yaml(stream=f)
        all_tests = load_model(content)
        return all_tests


def load_from_yaml(stream):
    add_constructor_for_custom_tags()
    return full_load(stream=stream)


def load_model(d: dict) -> AllTests:
    tests = [load_test(t) for t in d[key.TESTS]]
    return AllTests(
        module=d[key.MODULE],
        tests=tests,
    )


def load_test(d: dict) -> TestModel:
    description = load_test_description(d)
    target_func_name, target_class_name = load_target(d)
    init_parameters = load_named_values(d, key.INIT_PARAMETERS)
    parameters = load_named_values(d, key.PARAMETERS)
    mocks = load_mocks(d)
    patches = load_patches(d)
    exception_model = load_exception_model(d)
    result = load_result(d)
    return TestModel(
        name=load_required_name_key_value(d),
        short_name=load_short_name(d),
        description=description,
        target_func_name=target_func_name,
        target_class_name=target_class_name,
        init_parameters=init_parameters,
        parameters=parameters,
        result=result,
        exception=exception_model,
        mocks=mocks,
        patches=patches,
    )


def load_short_name(d: dict) -> str:
    return d[key.SHORT_NAME]


def load_result(d: dict) -> Any:
    return d.get(key.RESULT, None)


def load_target(d: dict) -> Tuple[str, str]:
    target = d[key.TARGET]
    return load_required_name_key_value(target), load_target_class_name(target)


def load_target_class_name(d: dict) -> str:
    return d.get(key.CLASS_NAME, "")


def load_test_description(d: dict) -> str:
    return d.get(key.DESCRIPTION, "")


def load_named_values(d: dict, key_name: str) -> NamedValues:
    """
    Load a NamedValues from YAML

    :param d:
    :param key_name:
    :return:
    """
    param_dict_list = d.get(key_name, [])
    name_value_list = [
        (load_optional_name_key_value(d), get_named_value_element_value(d))
        for d in param_dict_list
    ]
    return NamedValues(name_value_list)


def get_named_value_element_value(d: dict) -> str:
    return d[key.VALUE]


def load_mocks(d: dict) -> List[MockModel]:
    return [load_mock_model(m) for m in d.get(key.MOCKS, [])]


def load_mock_model(m: dict) -> MockModel:
    return MockModel(
        name=load_required_name_key_value(m),
        spec_str=m[key.SPEC],
        calls=load_mock_calls(m),
        attributes=load_mock_attributes(m),
    )


def load_mock_calls(m: dict) -> List[MockCallModel]:
    return [load_mock_call_model(c) for c in m.get(key.CALLS, [])]


def load_mock_attributes(m: dict) -> dict:
    return m.get(key.ATTRIBUTES, {})


def load_mock_call_model(c: dict) -> MockCallModel:
    return MockCallModel(
        name=load_optional_name_key_value(c),
        parameters=load_named_values(c, key.PARAMETERS),
        return_value=c.get(key.RETURN_VALUE, None),
    )


def load_patches(d: dict) -> List[PatchModel]:
    return [load_one_patch(p) for p in d.get(key.PATCHES, [])]


def load_one_patch(d: dict) -> PatchModel:
    return PatchModel(target=d[key.TARGET], replacement=d[key.PATCH_REPLACEMENT])


def load_exception_model(d: dict) -> Optional[ExceptionModel]:
    if key.EXCEPTION in d:
        return load_exception_internal(d[key.EXCEPTION])
    else:
        return None


def load_exception_internal(d: dict) -> ExceptionModel:
    return ExceptionModel(type=d[key.TYPE], message=d[key.MESSAGE])


def load_required_name_key_value(d: dict) -> str:
    return d[key.NAME]


def load_optional_name_key_value(d: dict) -> str:
    return d.get(key.NAME, "")



from __future__ import annotations

from distutils.util import strtobool
from typing import Any

from click import prompt

from testscribe.eval_expression import eval_expression
from testscribe.log import log
from testscribe.reflection_util import get_type_name
from testscribe.value_input_util import (
    transform_default_value,
    eval_special_value,
    get_string_value,
)
from testscribe.value_util import InputValue


def get_one_value_cli(prompt_name: str, t: type, default: Any):
    default_str = transform_default_value_cli(default=default, t=t)
    while True:
        try:
            return get_one_value_cli_internal(
                prompt_name=prompt_name, t=t, default=default_str
            )
        except Exception as e:
            msg = f"The value is invalid. Please try again.\nError detail:\n{str(e)}"
            log(msg)
            continue


def transform_default_value_cli(default: Any, t: type) -> str:
    default_str = transform_default_value(default=default, t=t)
    if isinstance(default, str) and t is not str:
        # Quote the string in case the type information is not available
        return repr(default_str)
    else:
        return default_str


def get_one_value_cli_internal(prompt_name: str, t: type, default: str):
    prompt_str = (
        f"Please provide the value for {prompt_name} of type: ({get_type_name(t)})"
    )
    raw_input_str = prompt(prompt_str, default=default, type=str)
    return post_process_raw_input(raw_input_str=raw_input_str, t=t)


def post_process_raw_input(raw_input_str: str, t: type) -> Any:
    eval_value = eval_special_value(raw_input_str)
    if eval_value is not None:
        return InputValue(expression=raw_input_str, value=eval_value)
    if t in {int, float, bool}:
        # noinspection PyTypeChecker
        return type_to_convert_func[t](raw_input_str)
    elif t is str:
        return get_string_value(raw_input_str)
    else:
        # assume the string is a valid python expression
        # ex. [1,2], ["a"], {"a": 2}, ('a', 1)
        return eval_expression(user_input=raw_input_str, t=t)


def convert_str_to_bool(s: str) -> bool:
    """
    True values are y, yes, t, true, on and 1; false values are n, no, f, false,
    off and 0.
    The input string is case-insensitive.
    Raises ValueError if s is anything else.

    :param s:
    :return:
    """
    return bool(strtobool(s))


type_to_convert_func = {int: int, float: float, bool: convert_str_to_bool}



from __future__ import annotations

from dataclasses import dataclass

from testscribe.base_reflection_util import get_class_instance_repr_with_full_name


@dataclass
class UserTriggeredException:
    """
    Used in an input expression to cause a mock call to throw the
    given exception.
    """

    exception: BaseException

    def __repr__(self):
        """
        Custom repr to allow easier code generation
        :return:
        """
        return get_class_instance_repr_with_full_name(self.exception)



import collections.abc
import typing
from inspect import Parameter, Signature
from types import (
    FunctionType,
    BuiltinFunctionType,
    MethodType,
    ModuleType,
)
from typing import (
    Any,
    Literal,
    AnyStr,
    NamedTuple,
    TypedDict,
    FrozenSet,
    DefaultDict,
    OrderedDict,
    ChainMap,
    Counter,
    Deque,
    Optional,
    Union,
)


def is_function_type(t: type):
    # doesn't include typing.Callable[[], int] or Callable
    return t in (FunctionType, BuiltinFunctionType, MethodType)


def is_function_instance(v: Any) -> bool:
    """
    Include method instances.

    :param v:
    :return:
    """
    return is_function_type(type(v))


def is_typing_callable_type(t: type):
    """
    Classes are callable. But this function only returns True
    for the Callable and its decorated form e.g. Callable[[int], int].
    :param t:
    :return:
    """
    return t in (typing.Callable, collections.abc.Callable) or get_type_origin(t) in (
        typing.Callable,
        collections.abc.Callable,
    )


def is_a_class_instance(v: Any) -> bool:
    return is_class_type(type(v))


def is_class_type(t: type) -> bool:
    if is_primitive_type(t) or is_container_type(t) or is_union_type(t):
        return False
    # Callable types are considered class types
    # Thus they can be mocked. e.g. a method can return Callable
    # todo: more checks for additional types
    # todo: find a more reliable way to test if it is a class type
    return True


def get_class_type(t: type) -> Optional[type]:
    """
    If the type is either a class type or a union type with at least one class type,
    returns the first class type.
    Return None otherwise.

    :param t:
    :return:
    """
    if is_class_type(t):
        return t
    if get_type_origin(t) == Union:
        element_types = get_type_args(t)
        for e_t in element_types:
            if is_class_type(e_t):
                return e_t
    return None


def is_primitive_type(t: type) -> bool:
    return t in [
        int,
        bool,
        str,
        float,
        bytes,
        type(None),
        ModuleType,
        Any,
        Parameter.empty,
        Signature.empty,
        Literal,
        AnyStr,
        # PyYaml has trouble saving the value Ellipsis
        # type(Ellipsis),
    ] or is_function_type(t)


def is_primitive_container_type(t: type):
    return t in [list, tuple, set, dict]


def is_container_type(t: type) -> bool:
    # todo: check for other container types
    # https://docs.python.org/3/library/typing.html
    return (
        is_primitive_container_type(t)
        or is_primitive_container_type(get_type_origin(t))
        or t
        in (
            NamedTuple,
            TypedDict,
            FrozenSet,
            DefaultDict,
            OrderedDict,
            ChainMap,
            Counter,
            Deque,
        )
    )


def get_type_origin(t: type):
    # https://docs.python.org/3/library/typing.html#typing.get_origin
    if hasattr(typing, "get_origin"):
        return typing.get_origin(t)
    elif hasattr(t, "__origin__"):  # pragma: no cover
        # https://www.python.org/dev/peps/pep-0585/
        return t.__origin__
    else:  # pragma: no cover
        return None


def get_type_args(t: type):
    # https://docs.python.org/3/library/typing.html#typing.get_origin
    if hasattr(typing, "get_args"):
        return typing.get_args(t)
    else:  # pragma: no cover
        # todo: work around for versions before python 3.8
        return None


def is_spec_value(value):
    return isinstance(value, type) or callable(value)


def is_union_type(t: type) -> bool:
    return get_type_origin(t) == Union



# Reflection related functions that don't depend on other modules in this project
from typing import Any

from testscribe.type_util import is_a_class_instance


def get_full_module_name(symbol: Any) -> str:
    if hasattr(symbol, "__module__"):
        return symbol.__module__
    else:
        return ""


def get_class_instance_repr_with_full_name(instance) -> str:
    assert is_a_class_instance(instance)
    full_module_name = get_full_module_name(instance)
    module_str = full_module_name + "." if full_module_name else ""
    return f"{module_str}{repr(instance)}"



from __future__ import annotations

from dataclasses import dataclass
from typing import Any


def get_value_repr(v) -> str:
    from testscribe.mock_proxy import MockProxy, get_proxy_str

    if isinstance(v, InputValue):
        return v.expression
    if isinstance(v, MockProxy):
        return get_proxy_str(v)
    return repr(v)


def get_value(v):
    if isinstance(v, InputValue):
        return v.value
    return v


@dataclass
class InputValue:
    """
    Encapsulate a value that is derived from an expression
    and the expression needs to be preserved to generate the correct
    test code.
    e.g.
    test_data.simple.C(1)
    ignore
    throw(Exception("err"))
    generate_data(1) generate_data is a user defined function
    that returns some complex values such as class instances.

    """

    # an expression to generate the value
    expression: str
    value: Any



from types import ModuleType
from typing import Any, Callable

from testscribe.model_type import ObjectModel, ModuleModel, SetModel
from testscribe.type_util import is_class_type, is_a_class_instance


def generic_is_special_type(
    value: Any, is_special_type_func: Callable[[Any], bool]
) -> bool:
    """

    :param is_special_type_func:
    :param value:
    :return: True if the value is a type that contains a type that will return
    True by the is_special_type_func check.
    """
    if is_special_type_func(value):
        return True
    elif is_enumerable_type(value):
        items = value
    elif isinstance(value, dict):
        items = value.values()
    else:
        return False

    for i in items:
        if generic_is_special_type(value=i, is_special_type_func=is_special_type_func):
            return True
    return False


def is_enumerable_type(value):
    return isinstance(value, list) or isinstance(value, tuple) or isinstance(value, set)


def is_complex_value(value):
    """
    Is this value of a type that need special assertions.

    :param value:
    :return:
    """
    t = type(value)
    return t in (ObjectModel, ModuleModel, SetModel)


def contain_complex_value(value):
    return generic_is_special_type(value=value, is_special_type_func=is_complex_value)


def has_unstable_repr_single_item(value) -> bool:
    """

    :param value: The value of a member of a class instance
    :return:
    """
    t = type(value)
    if callable(value) or t == ModuleType:
        return True
    if is_class_type(t):
        # If the member object doesn't implement __repr__,
        # it will be represented like
        # <test_data.simple.Dummy object at 0x7f55748d2be0>
        return not has_custom_repr_method(t) or object_contains_unstable_repr(value)
    return False


def has_unstable_repr(value) -> bool:
    return generic_is_special_type(
        value=value, is_special_type_func=has_unstable_repr_single_item
    )


def object_contains_unstable_repr(value) -> bool:
    """
    The value is an instance of a class. It contains a member that its repr is not
    stable i.e. it can change when run on different machines.
    e.g.
    a function's repr is like <function foo at 0x7fe688ff6b80>
    a module's repr is like
    <module 'test_data.simple' from '.../code/test-scribe/python/tests/test_data/simple.py'>
    a class instance's default repr is like <<test_data.simple.Dummy object
      at 0x7f55748d2be0>
    :param value:
    :return:
    """
    assert is_a_class_instance(value)
    if hasattr(value, "__dict__"):
        for name, v in vars(value).items():
            if has_unstable_repr(v):
                return True
    return False


def has_custom_repr_method(t: type) -> bool:
    """
    Return True if the type implements a custom __repr__ method.

    Return True also if the class inherit from a class that implements a
    custom __repr__ method. Ideally it should only return False in this case.
    There doesn't seem to be a reliable way to detect it though.
    """

    return t.__repr__ is not object.__repr__



from collections import defaultdict
from typing import List, Any

from testscribe.code_gen_util import (
    collect_return_values,
    is_same_value,
    join_lines,
)
from testscribe.ignore import IGNORED
from testscribe.model_type import MockModel, MockCallModel
from testscribe.user_triggered_exception import UserTriggeredException


def generate_property_init_statements(mock: MockModel) -> List[str]:
    return [
        f"{mock.name}.{name} = {repr(value)}" for name, value in mock.attributes.items()
    ]


def create_call_dict_indexed_by_method_name(mock_calls: List[MockCallModel]) -> dict:
    call_by_method_name = defaultdict(list)
    for call in mock_calls:
        method_name = call.name
        call_by_method_name[method_name].append(call)
    return call_by_method_name


def has_not_ignored_return_value(values):
    for v in values:
        if v is not IGNORED:
            return True
    return False


def gen_mock_return_values_for_one_method_with_return_values(
    mock_object_name: str, method_name: str, return_values: List[Any]
) -> str:
    if return_values:
        name = get_full_mock_method_name(method_name, mock_object_name)
        first_value = return_values[0]
        if is_single_real_return_value(return_values):
            return f"{name}.return_value = {repr(first_value)}"
        elif has_not_ignored_return_value(return_values):
            return f"{name}.side_effect = {repr(return_values)}"
    return ""


def is_single_real_return_value(return_values: List[Any]) -> bool:
    first_value = return_values[0]
    return (
        is_same_value(return_values)
        and not isinstance(first_value, UserTriggeredException)
        and first_value is not IGNORED
    )


def get_full_mock_method_name(method_name: str, mock_object_name: str) -> str:
    """
    Return the string used to identify the mock method.
    If the mock call is a call on the mock object itself the string will be
    the name of the mock object.

    :param method_name:
    :param mock_object_name:
    :return:
    """
    name = f"{mock_object_name}"
    if method_name:
        name += f".{method_name}"
    return name


def gen_mock_return_values_for_one_method(
    mock_object_name: str, method_name: str, methods: List[MockCallModel]
):
    return_values = collect_return_values(methods)
    return gen_mock_return_values_for_one_method_with_return_values(
        mock_object_name=mock_object_name,
        method_name=method_name,
        return_values=return_values,
    )


def generate_mock_return_values(mock: MockModel) -> List[str]:
    call_by_method_name = create_call_dict_indexed_by_method_name(mock.calls)
    # Assume empty elements in the list will be filtered out later
    return [
        gen_mock_return_values_for_one_method(
            mock_object_name=mock.name, method_name=method_name, methods=methods
        )
        for method_name, methods in call_by_method_name.items()
    ]


def generate_one_mock_creation_statement(mock: MockModel) -> str:
    spec = mock.spec_str
    # todo: add an option to allow spec_set=True to be set for stronger checks
    # see
    # https://docs.python.org/3/library/unittest.mock.html#unittest.mock.create_autospec
    return f"{mock.name}: {spec} = create_autospec(spec={spec})"


def generate_behavior_statements_for_one_mock(mock: MockModel) -> List[str]:
    property_init_statements = generate_property_init_statements(mock)
    return_value_statements = generate_mock_return_values(mock)
    return property_init_statements + return_value_statements


def generate_mock_behavior_statements(mocks: List[MockModel]) -> List[str]:
    statements = []
    for m in mocks:
        statements_for_one_mock = generate_behavior_statements_for_one_mock(m)
        statements.extend(statements_for_one_mock)
    return statements


def generate_mocks_str(mocks: List[MockModel]) -> str:
    # Create all the mock creation statements first
    # so that they can be freely referenced later including as return values.
    mock_creation_statements = [generate_one_mock_creation_statement(m) for m in mocks]
    mock_behavior_statements = generate_mock_behavior_statements(mocks)
    return join_lines(
        mock_creation_statements + mock_behavior_statements,
        prepend_new_line=True,
        indentation_level=1,
    )



from __future__ import annotations

import itertools
import logging
import re
from hashlib import md5
from importlib import import_module
from typing import List, Any

logger = logging.getLogger(__name__)

BUILTIN_MODULE_NAME = "builtins"


def generic_transform(v: Any, transform_func: callable) -> Any:
    def _transform(value: Any) -> Any:
        return generic_transform(v=value, transform_func=transform_func)

    # todo: transform the member values if it is a class.
    if isinstance(v, list):
        new_v = [_transform(item) for item in v]
    elif isinstance(v, tuple):
        new_v = tuple([_transform(item) for item in v])
    elif isinstance(v, set):
        new_v = {_transform(item) for item in v}
    elif isinstance(v, dict):
        new_v = {_transform(k): _transform(value) for k, value in v.items()}
    else:
        new_v = v
    return transform_func(new_v)


pattern = re.compile(r"(?<!^)(?=[A-Z])")


def convert_camel_case_to_snake_case(s: str):
    # adopted from
    # https://stackoverflow.com/questions/1175208/elegant-python-function-to-convert-camelcase-to-snake-case
    snake_case_str = pattern.sub("_", s).lower()
    return snake_case_str


TRAILING_NUMBERS_PATTERN = re.compile(r"_\d+$")


def remove_trailing_numbers(s: str) -> str:
    return TRAILING_NUMBERS_PATTERN.sub("", s)


def flattern_list(list_of_list: List[List]):
    """
    It doesn't recursively flattern lists. e.g. input [[["a"]]]
    will result in [["a"]].
    :param list_of_list:
    :return:
    """
    return list(itertools.chain(*list_of_list))


def consistent_hash_str(s: str) -> int:
    return int(md5(s.encode()).hexdigest(), 16)


def load_object(symbol_name: str, module_str: str) -> Any:
    module_obj = import_module(module_str)
    obj = getattr(module_obj, symbol_name, None)
    return obj



from __future__ import annotations

import logging
import sys
from dataclasses import dataclass
from pathlib import Path
from typing import Callable, Union

from yaml import safe_load

from testscribe import global_var
from testscribe.api.io_provider import IOProvider
from testscribe.error import Error
from testscribe.file_info import get_function
from testscribe.reflection_util import get_module_and_symbol, get_symbol

g_indent = "    "

logger = logging.getLogger(__name__)


@dataclass(frozen=True)
class Config:
    output_root_path: Path
    setup_function: Callable


CONFIG_FILE_NAME = "test-scribe-config.yml"
ADDITIONAL_PYTHON_PATH_KEY = "python-paths"
OUTPUT_ROOT_DIR_KEY = "output-root-dir"
SETUP_FUNCTION_KEY = "setup-function"
IO_PROVIDER_CLASS_NAME = "io-provider-full-class-name"


def get_setup_func(data: dict) -> Union[Callable, None]:
    if SETUP_FUNCTION_KEY not in data:
        return None
    full_name = data[SETUP_FUNCTION_KEY]
    base_msg = (
        f"The setup function as specified by the {SETUP_FUNCTION_KEY} key is:"
        f"{full_name}. It should be a valid fully qualified function name."
    )
    if not isinstance(full_name, str) or full_name == "":
        raise Error(f"{base_msg} The value is not a string or is an empty string.")
    try:
        module_str, function_name = get_module_and_symbol(full_name)
        func, clazz = get_function(module_name=module_str, func_name=function_name)
    except Exception as e:
        raise Error(f"{base_msg} Can't load this symbol.\nerror detail:\n{e}")
    # get_module_and_symbol assumes the name represents a function.
    # It can't handle a method name e.g. test_data.simple.C.bar
    assert clazz is None
    return func


def add_additional_python_paths(data: dict):
    path_strings = data.get(ADDITIONAL_PYTHON_PATH_KEY, [])
    sys.path.extend(path_strings)


def get_output_root_path(data: dict) -> Path:
    output_root_dir = data.get(OUTPUT_ROOT_DIR_KEY, "")
    if output_root_dir:
        return Path(output_root_dir)
    else:
        return Path()


def initialize_io(data: dict) -> IOProvider:
    io_provider_class_name = data.get(IO_PROVIDER_CLASS_NAME, "")
    if io_provider_class_name:
        io_provider_class = get_symbol(io_provider_class_name)
        global_var.g_io = io_provider_class()
    return global_var.g_io


def init_config(config_file_path: Path) -> Config:
    # todo: make io provider configurable
    data = load_config_data(config_file_path)
    add_additional_python_paths(data)
    initialize_io(data)
    output_root_path = get_output_root_path(data)
    func = get_setup_func(data)
    return Config(
        output_root_path=output_root_path,
        setup_function=func,
    )


def load_config_data(config_file_path: Path) -> dict:
    data = {}
    if config_file_path.exists():
        with open(config_file_path) as file:
            data = safe_load(file)
    return data



from __future__ import annotations

import logging
from functools import partial
from inspect import signature
from typing import Union, Any, Callable, Optional
from unittest.mock import Mock, call

from testscribe import global_var
from testscribe.custom_type import Spec
from testscribe.error import Error
from testscribe.mock_proxy import MockProxy, is_mock_proxy
from testscribe.patcher import patch_common, remove_duplicate_patch
from testscribe.reflection_util import (
    has_param_names,
    get_bound_arguments,
    get_symbol,
    get_method_signature_for_caller,
    get_full_spec_name,
)
from testscribe.type_util import is_spec_value

logger = logging.getLogger(__name__)


def normalize_mock_call(mock_call, spec: Spec):
    """
    Return a tuple representing a mock call with the keyword arguments only.

    :param mock_call: a mock call of a Mock object
    :param spec: the spec of the Mock object
    :return: a tuple
        name: method name if this is a method call or "" if this is a call on the
        object itself
        args: should be ()
        kargs: keyword arguments dictionary
    """
    name, args, kwargs = mock_call
    get_bound_arguments_partial = partial(get_bound_arguments, args=args, kwargs=kwargs)
    if name:
        # this is a method call
        # todo: handle the error that the method doesn't exist.
        # The error should have been caught error when the method is
        # invoked.
        method_sig = get_method_signature_for_caller(clazz=spec, name=name)
        args = get_bound_arguments_partial(sig=method_sig)
        return getattr(call, name)(**args)
    else:
        # todo: handle __call__
        # If the parameter names are not available, it is simper to compare
        # directly with mock.mock_calls instead of using this function.
        assert has_param_names(spec)
        args = get_bound_arguments_partial(sig=signature(spec))
        return call(**args)


def get_direct_mock_calls(mock: Mock) -> list:
    # filter out the calls on the returned instances (child mocks).
    return [c for c in mock.mock_calls if "()" not in c[0]]


def get_normalized_mock_calls_internal(mock: Mock, spec: Spec) -> list:
    # The calls on the returned instances (child mocks) are asserted separately.
    direct_calls = get_direct_mock_calls(mock)
    actual_calls = [normalize_mock_call(mock_call=c, spec=spec) for c in direct_calls]
    return actual_calls


def get_target_str_from_obj(obj: Union[type, Callable, MockProxy]) -> str:
    """
    :param obj: An object to be patched e.g. class, function
    :return: The fully qualified name for the object
    """
    check_target_can_be_used_as_spec(obj)
    # The obj may have been patched.
    obj_source = obj.spec_test_scribe_ if is_mock_proxy(obj) else obj
    return get_full_spec_name(obj_source)


def check_target_can_be_used_as_spec(obj):
    if not is_spec_value(obj):
        msg = (
            f"{obj} is not a type or Callable."
            " It can't be used as a spec for a mock object."
        )
        raise Error(msg)


def get_mock_target_str(target: Union[str, type, Callable, MockProxy]) -> str:
    if isinstance(target, str):
        return target
    else:
        return get_target_str_from_obj(target)


def patch_with_mock_internal(
    target: Union[str, type, Callable, MockProxy], mock_name: str, spec: Optional[Spec]
) -> None:
    """
    A convenience function to make patching with a mock object easier.

    :param spec: If None, infer the spec from the target
    :param mock_name:
    :param target: If it is not a string, it should be the symbol that can to be
    patched with a mock object. In this case, the spec for the mock object
    and the target have to match. If they don't e.g. a function mod_a.foo is imported
    into mod_b as foo and invoked as foo, to mock the invocation, use the string
    "mod_b.foo" as the value for the target parameter.
    It's possible that the target has been patched, in this case the target is an
    instance of MockProxy.
    """
    if not global_var.g_test_generating_mode:
        logger.debug("This is not a test generation session. Ignore patch.")
        return
    target_str = get_mock_target_str(target)
    # has to remove duplicate patches first before creating a mock
    # to avoid interferance of the mock name creation.
    remove_duplicate_patch(target_str)
    if spec is None:
        obj = get_symbol(target_str)
        check_target_can_be_used_as_spec(obj)
        spec = obj
    mock_obj = MockProxy(spec=spec, name=mock_name)
    patch_common(target_str=target_str, replacement=mock_obj)


def patch_with_expression_internal(target_str: str, expression: str) -> None:
    # It's possible to derive the type so that the m and c convenience symbols
    # can be used without specifying spec types.
    # However, it may not be necessary given the patch_with_mock function.
    if not global_var.g_test_generating_mode:
        logger.debug("This is not a test generation session. Ignore patch.")
        return
    remove_duplicate_patch(target_str)

    # avoid circular dependencies
    from testscribe.eval_expression import eval_expression

    replacement = eval_expression(user_input=expression, t=Any)
    patch_common(target_str=target_str, replacement=replacement)



from inspect import Parameter
from pathlib import Path
from typing import Any, List

from testscribe.context import Context
from testscribe.namedvalues import NamedValues


class IOProvider:
    """
    The input and output plugin module implementations need to support these methods.
    """

    def __init__(self):
        pass

    def end(self, scribe_file_path: Path):
        pass

    def log(self, s: str) -> None:
        raise NotImplementedError()

    def get_parameter_value(
        self, param_info_list: List[Parameter], defaults: list, context: Context
    ) -> NamedValues:
        """

        :param param_info_list: the list should not be empty, the no parameter case
        should have been handled by the caller.
        :param defaults: corresponding default values for the parameters
        :param context:
        :return:
        """
        raise NotImplementedError()

    def get_one_value(
        self, prompt_name: str, name: str, t: type, default: Any, context: Context
    ) -> Any:
        raise NotImplementedError()

    def get_test_description(self, default: str) -> str:
        raise NotImplementedError()

    def get_short_test_name(self, default_short_name: str) -> str:
        raise NotImplementedError()



from __future__ import annotations

import logging
import sys
from importlib import import_module
from inspect import isfunction, getmembers, isclass
from pathlib import Path
from typing import List, Tuple, Union

from testscribe.error import Error
from testscribe.module import Module
from testscribe.util import load_object

logger = logging.getLogger(__name__)


def get_module(target_file: Path) -> Module:
    """
    Walk up the target file path to find the first parent that is in the sys.path.

    :param target_file:
    :return:
    """
    absolute_sys_paths = [Path(s).absolute() for s in sys.path]

    def get_module_name(dir_path: Path) -> List[str]:
        if dir_path.name == "":
            raise Error(f"The file ({target_file}) can't be loaded as a python module.")

        if dir_path in absolute_sys_paths:
            return []
        parent_path = dir_path.parent
        names = get_module_name(parent_path)
        names.append(dir_path.name)
        return names

    module_names = get_module_name(target_file.parent)
    module_names.append(target_file.stem)
    return Module(module_names)


def get_function(
    module_name: str, func_name: str
) -> Tuple[callable, Union[type, None]]:
    """

    :param module_name:
    :param func_name:
    :return: A tuple of the function object and class type if this is a method
    """
    obj = load_object(symbol_name=func_name, module_str=module_name)
    if isfunction(obj):
        return obj, None

    # todo: add a class parameter to explicitly indicate the target class
    # in a rare case when the target function is defined in two classes in the
    # same file.
    mod = import_module(module_name)
    classes = getmembers(mod, isclass)
    for class_name, class_type in classes:
        methods = getmembers(class_type, isfunction)
        for method_name, method_obj in methods:
            if method_name == func_name:
                return method_obj, class_type
    raise Error(
        f"Can't find the function or method with the name {func_name} in module {module_name}."
    )



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Optional

from testscribe.namedvalues import NamedValues


@dataclass(frozen=True)
class RuntimeValues:
    # parameters to the constructor if the target is a method
    init_values: NamedValues
    instance: Any  # class instance
    # The init_values and values have been transformed to a format suitable to be
    # saved in a YAML format.
    values: Optional[NamedValues]
    # The transformed result
    # The result is None if there is an exception.
    result: Any
    exception: Exception



from __future__ import annotations

from typing import Collection, List


class Module:
    def __init__(self, module_names: Collection[str]):
        """

        :param module_names:a list of string representing the full module name.
        Parent module name comes first.
        """
        self.names = tuple(module_names)

    def get_package_name_list(self) -> List[str]:
        return list(self.names[:-1])

    def get_module_name_only(self) -> str:
        return self.names[-1]

    def get_module_str(self):
        return ".".join(self.names)


def get_module_from_str(name: str) -> Module:
    return Module(name.split("."))



class SingleLineString:
    """
    A special type to represent a string type that should only contain one line
    of text.
    """

    pass


class NoDefault:
    """
    Marker object for not having a default value.
    Note that None can be a valid value.
    """



from __future__ import annotations

import logging
from typing import Tuple, Any, Iterable

from testscribe.value_util import get_value_repr

logger = logging.getLogger(__name__)


class NameNotFound:
    """
    Marker object for not having the name in the NamedValues instance.
    Note that None can be a valid value.
    """


class NamedValues:
    def __init__(self, name_value_list: Iterable[Tuple[str, Any]] = []):
        self.params = list(name_value_list)

    def __getitem__(self, index: int) -> Tuple[str, Any]:
        """
        Support getting an item via its index
        e.g. see testscribe.value_input.get_default_value_from_old_params
        """
        return self.params[index]

    def get_size(self):
        return len(self.params)

    def get_value_by_name(self, name: str):
        for n, value in self.params:
            if n == name:
                return value
        return NameNotFound

    def __repr__(self):
        return f"NamedValues({repr(self.params)})"

    def as_arg_str(self):
        # Some functions e.g. some builtin functions may not have the parameter
        # names available.
        arg_list = [format_one_param(name=k, value=v) for k, v in self.params]
        arg_str = ", ".join(arg_list)
        return arg_str

    def as_list(self):
        return list(self.params)


def format_one_param(name: str, value: Any) -> str:
    value_repr = get_value_repr(value)
    if name:
        return f"{name}={value_repr}"
    else:
        return value_repr



from __future__ import annotations

import functools
import traceback

from click import Abort

from testscribe.execution_util import ERROR_RETURN_CODE
from testscribe.log import log


def get_exception_detail_string(e):
    return "".join(traceback.format_exception(type(e), e, e.__traceback__))


def log_exception(e: BaseException):
    exception_msg = get_exception_detail_string(e)
    msg = f"Aborted due to an exception:\n{exception_msg}"
    log(msg)
    raise Abort(ERROR_RETURN_CODE)


def exception_handler(func):
    @functools.wraps(func)
    def wrapper_decorator(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except BaseException as e:
            return log_exception(e)

    return wrapper_decorator



from __future__ import annotations

import logging
from collections.abc import Callable
from typing import List, get_type_hints, Any, Optional

from testscribe import global_var
from testscribe.context import Context
from testscribe.custom_type import Spec
from testscribe.execution_util import show_user_call_stack
from testscribe.log import log
from testscribe.mock_call import MockCall
from testscribe.model_type import TestModel, MockModel, get_mock_by_name
from testscribe.special_type import NoDefault
from testscribe.type_util import get_type_origin
from testscribe.util import convert_camel_case_to_snake_case

UNSUPPORTED_ATTRIBUTES = {"__len__"}

logger = logging.getLogger(__name__)


def create_mock_name(name: str, spec: Spec) -> str:
    base_name = create_default_mock_name(spec) if name == "" else name
    return create_unique_mock_name(base_name)


def create_default_mock_name(spec: Spec) -> str:
    type_name_in_camel_case = convert_camel_case_to_snake_case(get_spec_name(spec))
    return f"m_{type_name_in_camel_case}"


def get_spec_name(spec: Spec) -> str:
    type_origin = get_type_origin(spec)
    if type_origin is Callable:
        # Callable doesn't have __qualname__ nor __name__ attribute
        return "Callable"
    else:
        # todo: is it possible that __qualname__ is not available?
        return spec.__qualname__


def create_unique_mock_name(name: str) -> str:
    counter = global_var.g_mock_name_counter
    unique_name = f"{name}_{counter[name]}" if name in counter else name
    counter[name] += 1
    return unique_name


def check_unsupported_attributes(attribute_name: str):
    # Debuggers call hasattr("<unsupported method name") which results in
    # these calls which are not part of the program's logic.
    if attribute_name in UNSUPPORTED_ATTRIBUTES:
        raise AttributeError(f"mocking the {attribute_name} method is not supported.")


def create_mock_call(
    method_name: str, mock_name: str, spec: Spec, mock_calls: List[MockCall]
) -> MockCall:
    logger.debug(f"Creating a new mock call ({method_name})")
    previous_call_count = get_previous_call_count(
        mock_calls=mock_calls, method_name=method_name
    )
    method = MockCall(
        method_name=method_name,
        mock_name=mock_name,
        spec=spec,
        previous_call_count=previous_call_count,
    )
    return method


def get_previous_call_count(mock_calls: List[MockCall], method_name: str) -> int:
    # how many times has this method been called
    count = 0
    for call in mock_calls:
        if call.method_name == method_name:
            count += 1
    return count


def get_attribute_type(spec: Spec, name: str) -> type:
    t = get_type_from_type_hints(spec, name)
    if t is None:
        if name in dir(spec):
            # type hints method below doesn't work for methods
            attrib = getattr(spec, name)
            t = type(attrib)
        else:
            t = Any
    return t


def get_type_from_type_hints(spec: Spec, attribute_name: str) -> Optional[type]:
    """
    Get type information from attribute's type annotaiton.

    dataclass for example doesn't seem to have a way to define class variables
    for the purpose of providing type information for a field.
    e.g.
    field1 = 5 will create a class variable but not a field.

    field1: int = 5 will not create a class variable
    get_type_hints(data class) will return {'field1': <class 'int'>} in this case.

    get_type_hints(a function) will return type annotations for the
    parameters and return value
    e.g. {'p': <class 'test_data.person.Person'>, 'return': <class 'str'>}

    """
    hints = get_type_hints(spec)
    return hints.get(attribute_name, None)


def get_default_mock_attribute_value(
    mock_name: str,
    attribute_name: str,
    test_to_infer_default_inputs: Optional[TestModel],
) -> Any:
    if test_to_infer_default_inputs:
        existing_mock = get_mock_by_name(
            mocks=test_to_infer_default_inputs.mocks, name=mock_name
        )
        return infer_default_mock_attribute_value_from_mock(
            attribute_name=attribute_name, existing_mock=existing_mock
        )
    return NoDefault


def infer_default_mock_attribute_value_from_mock(
    attribute_name: str, existing_mock: Optional[MockModel]
) -> Any:
    if existing_mock:
        return existing_mock.attributes.get(attribute_name, NoDefault)
    else:
        return NoDefault


def get_mock_attribute_value(attribute_name: str, mock_name: str, spec: Spec):
    # todo: create a strict mode to check the spec for valid
    #  attributes.
    # Many attributes are created dynamically, so it is error
    # prone to check by default. see
    # https://docs.python.org/3/library/unittest.mock.html#auto-speccing
    log(
        f"Mock object {mock_name}'s ( {attribute_name} )"
        " attribute is accessed for the first time."
    )
    # 3 frames between user code and show_user_call_stack
    # __getattr__, record_attribute, get_mock_attribute_value
    # if the call stack changes, adjust the parameter here.
    show_user_call_stack(additional_num_system_frame=3)
    description = f"Mock object {mock_name}'s {attribute_name} attribute"
    t = get_attribute_type(spec=spec, name=attribute_name)
    default = get_default_mock_attribute_value(
        mock_name=mock_name,
        attribute_name=attribute_name,
        test_to_infer_default_inputs=global_var.g_test_to_infer_default_inputs,
    )
    # avoid circular dependency
    from testscribe.value_input import get_one_value

    input_value = get_one_value(
        prompt_name=f"the {attribute_name} attribute",
        t=t,
        context=Context(description),
        default=default,
    )
    return input_value



from __future__ import annotations

import logging
from typing import Any

from testscribe import global_var
from testscribe.context import Context

logger = logging.getLogger(__name__)


def get_one_value(
    prompt_name: str, t: type, context: Context, default: Any, name: str = ""
) -> Any:
    """

    :param prompt_name: the string used as a prompt for the input
    :param t:
    :param context:
    :param default:
    :param name: a name associated with the value e.g. a parameter's name
    :return:
    """
    logger.debug(
        f"Get one value. prompt_name: {prompt_name}, name: {name}, "
        f"type: {t}, context: {context}, default: {default} "
    )
    # signature returns None, get_type_hints returns <class 'NoneType'>
    if t in [None, type(None)]:
        return None
    return global_var.g_io.get_one_value(
        prompt_name=prompt_name, name=name, t=t, default=default, context=context
    )



from __future__ import annotations

from collections import Counter
from dataclasses import dataclass
from typing import Any
from unittest.mock import patch

from testscribe import global_var
from testscribe.log import log
from testscribe.model_type import MockNameModel
from testscribe.transformer import transform_value
from testscribe.util import remove_trailing_numbers
from testscribe.value_util import get_value, get_value_repr


def patch_common(target_str: str, replacement: Any):
    """
    :param target_str:
    :param replacement:
    :return:
    """
    patchers = global_var.g_patchers
    value = get_value(replacement)
    value_str = get_value_repr(value)
    log(f"Patch {target_str} with {value_str}")
    instance = patch(target=target_str, new=value)
    patchers[target_str] = create_patcher(
        target_str=target_str, instance=instance, replacement=replacement
    )
    instance.start()


def create_patcher(target_str: str, instance, replacement: Any) -> Patcher:
    """

    :param target_str: the target symbol to patch
    :param instance: the return value of a mock.patch call. The type _patch is
    an internal type of mock.
    :param replacement:
    :return:
    """
    # Need to take a snapshot of the value
    # so that the generated code will use the initial value not the potentially
    # mutated value. e.g. if the replacement is a list, it can potentially
    # change its value after execution.

    # Transform process is itself a deepcopy.

    # Wrapping the value as an ExpressionModel doesn't work when the value
    # contains a MockProxy
    return Patcher(
        target=target_str,
        instance=instance,
        replacement_spec=transform_value(replacement),
    )


@dataclass
class Patcher:
    target: str
    # the return value of a mock.patch call. The type _patch is
    #     an internal type of mock. It can't be used directly here.
    instance: Any
    # A transformed copy of the origianl replacement value
    replacement_spec: Any


def remove_duplicate_patch(target_str: str):
    patchers = global_var.g_patchers
    if target_str in patchers:
        log(f"Remove previous patch of the same target {target_str}")
        patcher = patchers.pop(target_str)
        patcher.instance.stop()
        remove_mock(patcher.replacement_spec)


def remove_mock(replacement_spec: Any) -> None:
    if isinstance(replacement_spec, MockNameModel):
        name = replacement_spec.name
        assert name in global_var.g_name_mock_dict
        del global_var.g_name_mock_dict[name]
        remove_mock_name(name)


def remove_mock_name(name: str):
    nomalized_name = remove_trailing_numbers(name)
    decrement_counter(
        name_counter=global_var.g_mock_name_counter, nomalized_name=nomalized_name
    )


def decrement_counter(name_counter: Counter, nomalized_name: str) -> Counter:
    assert nomalized_name in name_counter
    name_counter[nomalized_name] -= 1
    if name_counter[nomalized_name] == 0:
        del name_counter[nomalized_name]
    return name_counter



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['testscribe', 'testscribe.api']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.2', 'typer']

setup_kwargs = {
    'name': 'testscribe',
    'version': '0.0.1a1',
    'description': 'Unit test automation tool',
    'long_description': '# TestScribe for Python\n\nUnit test made easy\n\nA tool to make python unit testing easier by automating the boring and repetitive parts.\n\n',
    'author': 'Ray Yang',
    'author_email': 'ruiguo@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



from __future__ import annotations

from inspect import Parameter
from typing import List

from testscribe.context import Context
from testscribe.namedvalues import NamedValues
from testscribe.value_input import get_one_value


def get_parameter_value_cli(
    param_info_list: List[Parameter], defaults: list
) -> NamedValues:
    params = []
    for param, default_value in zip(param_info_list, defaults):
        v = get_one_value(
            prompt_name=f"the parameter ({param.name})",
            name=param.name,
            t=param.annotation,
            context=Context(""),
            default=default_value,
        )
        params.append((param.name, v))
    # log(f"Params:{params}")
    return NamedValues(params)



from __future__ import annotations

from testscribe import global_var
from testscribe.model_type import AllTests


def get_test_description(
    all_tests: AllTests, index_of_test_to_update: int, ask_for_description: bool
) -> str:
    default_test_description = get_default_description(
        all_tests=all_tests, index_of_test_to_update=index_of_test_to_update
    )
    if ask_for_description:
        return global_var.g_io.get_test_description(default_test_description)
    else:
        return default_test_description


def get_default_description(all_tests: AllTests, index_of_test_to_update: int):
    if index_of_test_to_update >= 0:
        return all_tests.tests[index_of_test_to_update].description
    else:
        return ""
